
# coding: utf-8

# In[1]:

import numpy as np
import math
import PIL
import transforms3d
from numpy import cos,sin

# In[36]:

def qmul(q1, q2):
    a1, b1, c1, d1 = q1
    a2, b2, c2, d2 = q2
    r1 = a1*a2 - b1*b2 - c1*c2 - d1*d2
    r2 = a1*b2 + a2*b1 + c1*d2 - c2*d1
    r3 = a1*c2 + a2*c1 - b1*d2 + b2*d1
    r4 = a1*d2 + a2*d1 + b1*c2 - b2*c1
    return np.array([r1, r2, r3, r4]).astype(np.float32)


# In[73]:

def qinv(q):
    a, b, c, d = q
    norm2 = float(a**2 + b**2 + c**2 + d**2)
    return np.array([a, -b, -c, -d]).astype(np.float32)/(norm2 + 1e-10)


# In[74]:

def qtrans(q1, q2):
    #q1 / q2 = q1 * q2^-1
    return qmul(qinv(q2), q1)


# In[75]:

def ptrans(p1, p2):
    return p1 - p2


# In[76]:

def trans(z1, z2):
    p = ptrans(z1[:3], z2[:3])
    q = qtrans(z1[3:], z2[3:])
    return np.concatenate([p,q],0)


def to_r(q):
    a,b,c,d = q
    R = np.array([[a**2 + b**2 - c**2 - d**2, 2*b*c - 2*a*d, 2*b*d + 2*a*c],
                 [2*b*c + 2*a*d, a**2-b**2+c**2-d**2, 2*c*d - 2*a*b],
                 [2*b*d - 2*a*c, 2*c*d + 2*a*b, a**2 - b**2 - c**2 + d**2]], dtype=np.float32)
    return R



def isRotationMatrix(R) :
    Rt = np.transpose(R)
    shouldBeIdentity = np.dot(Rt, R)
    I = np.identity(3, dtype = R.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-2


# Calculates rotation matrix to euler angles
# The result is the same as MATLAB except the order
# of the euler angles ( x and z are swapped ).
def rotationMatrixToEulerAngles(R) :

    assert(isRotationMatrix(R))

    sy = math.sqrt(R[0,0] * R[0,0] +  R[1,0] * R[1,0])

    singular = sy < 1e-6

    if  not singular :
        x = math.atan2(R[2,1] , R[2,2])
        y = math.atan2(-R[2,0], sy)
        z = math.atan2(R[1,0], R[0,0])
    else :
        x = math.atan2(-R[1,2], R[1,1])
        y = math.atan2(-R[2,0], sy)
        z = 0

    return np.array([x, y, z])

def trans2(z1, z2):
    R1 = to_r(z1[3:])
    T1 = -np.dot(R1, np.array(z1[:3]))
    R2 = to_r(z2[3:])
    T2 = -np.dot(R2, np.array(z2[:3]))

    RT1 = np.concatenate([R1, np.expand_dims(T1,1)], 1);
    RT2 = np.concatenate([R2, np.expand_dims(T2,1)], 1);

    e =  np.array([[0,0,0,1]])

    M1 = np.concatenate([RT1, e], 0)
    M2 = np.concatenate([RT2, e], 0)
    #print(M1, M2)

    dM = np.dot(M1, np.linalg.inv(M2))
    #dR = np.dot(R2.transpose(), R1)
    dR = rotationMatrixToEulerAngles(dM[:3, :3])
    #dT = dM[:3, -1]
    #dT = np.dot(R2.transpose(), T1 - T2)
    dT = dM[:3, -1]
    return (dT, dR)

def transfromM(dM):
    dR = rotationMatrixToEulerAngles(dM[:3, :3])
    dT = dM[:3, -1]
    return (dT, dR)




def rotateImage(img, angle):
    img = np.asarray(img)
    width = img.shape[1]
    angle = angle/np.pi

    if angle < 0:
        angle += 2

    if angle >= 0:
        shift = int(angle * width / 2)
        img = np.concatenate([img[:,shift:,:], img[:, :shift, :]], 1)

    return PIL.Image.fromarray(img)


def generate_transformation_matrix(x,y,z,yaw,pitch,roll):
    current_t = np.eye(4)
    current_t[0,-1] = x
    current_t[1,-1] = y
    current_t[2,-1] = z       
    alpha = yaw
    beta = pitch
    gamma = roll
    cpose = np.zeros(16) 
    cpose[0] = cos(alpha) * cos(beta);
    cpose[1] = cos(alpha) * sin(beta) * sin(gamma) - sin(alpha) * cos(gamma);
    cpose[2] = cos(alpha) * sin(beta) * cos(gamma) + sin(alpha) * sin(gamma);
    cpose[3] = 0
    cpose[4] = sin(alpha) * cos(beta);
    cpose[5] = sin(alpha) * sin(beta) * sin(gamma) + cos(alpha) * cos(gamma);
    cpose[6] = sin(alpha) * sin(beta) * cos(gamma) - cos(alpha) * sin(gamma);
    cpose[7] = 0
    cpose[8] = -sin(beta);
    cpose[9] = cos(beta) * sin(gamma);
    cpose[10] = cos(beta) * cos(gamma);
    cpose[11] = 0
    cpose[12:16] = 0
    cpose[15] = 1           
    cpose = cpose.reshape((4,4))      
    cpose = np.dot(cpose, current_t)
    current_rt = cpose
    rotation = np.array([[0,-1,0,0],[-1,0,0,0],[0,0,1,0],[0,0,0,1]])
    current_rt = np.dot(rotation, current_rt)
    return current_rt




def mat_to_str(matrix):
    s = ""
    for row in range(4):
        for col in range(4):
            s = s + " " + str(matrix[row][col])
    return s.strip()


def mat_to_posi_xyz(cpose):
    return cpose[:3, -1]

def mat_to_quat_xyzw(cpose):
    rot = cpose[:3, :3]
    ## Return: [r_x, r_y, r_z, r_w]
    wxyz = transforms3d.quaternions.mat2quat(rot)
    return quat_wxyz_to_xyzw(wxyz)



## Blender generated poses and OpenGL default poses adhere to
## different conventions. To better understand the nitty-gritty
## transformations inside this file, check out this:
## https://en.wikibooks.org/wiki/OpenGL_Programming/Modern_OpenGL_Tutorial_Navigation
##  Blender: z-is-up
##      same with: csv, rt_camera_matrix
##  OpenGL: y-is-up
##      same with: obj
##  Default camera: y is up-direction, -z facing


## Quat(wxyz)
def quat_pos_to_mat(pos, quat):
    r_w, r_x, r_y, r_z = quat
    #print("quat", r_w, r_x, r_y, r_z)
    mat = np.eye(4)
    mat[:3, :3] = transforms3d.quaternions.quat2mat([r_w, r_x, r_y, r_z])
    mat[:3, -1] = pos
    # Return: roll, pitch, yaw
    return mat

## Used for URDF models that are default -x facing
##  Rotate the model around its internal x axis for 90 degrees
##  so that it is at "normal" pose when applied camera_rt_matrix 
## Format: wxyz for input & return
def z_up_to_y_up(quat_wxyz):
    ## Operations (1) rotate around y for pi/2, 
    ##            (2) rotate around z for pi/2
    to_y_up = transforms3d.euler.euler2quat(0, np.pi/2, np.pi/2)
    return transforms3d.quaternions.qmult(quat_wxyz, to_y_up)

## Models coming out of opengl are negative x facing
##  Transform the default to 
def y_up_to_z_up(quat_wxyz):
    to_z_up = transforms3d.euler.euler2quat(np.pi/2, 0, np.pi/2)
    return transforms3d.quaternions.qmult(to_z_up, quat_wxyz)


def quat_wxyz_to_euler(wxyz):
    q0, q1, q2, q3 = wxyz
    sinr = 2 * (q0 * q1 + q2 * q3)
    cosr = 1 - 2 * (q1 * q1 + q2 * q2)
    sinp = 2 * (q0 * q2 - q3 * q1)
    siny = 2 * (q0 * q3 + q1 * q2)
    cosy = 1 - 2 * (q2 * q2 + q3 * q3)

    roll  = np.arctan2(sinr, cosr)
    pitch = np.arcsin(sinp)
    yaw   = np.arctan2(siny, cosy)
    return [roll, pitch, yaw]


## wxyz: numpy array format
def quat_wxyz_to_xyzw(wxyz):
    return np.concatenate((wxyz[1:], wxyz[:1]))

## xyzw: numpy array format
def quat_xyzw_to_wxyz(xyzw):
    return np.concatenate((xyzw[-1:], xyzw[:-1]))


from gibson.assets.assets_manager import AssetsManager


def set_assets_path():
    assets_manager = AssetsManager()
    print('Current assets path = ' + assets_manager.get_assets_path())
    path = input('Insert the new assets path\n')
    assets_manager.set_assets_path(path).save_assets_information()
    print('Assets path correctly changed!')


#!/usr/bin/env python3
# -*- Coding: UTF-8 -*-

# ---------------------------------------------------------------------------
# Open Asset Import Library (ASSIMP)
# ---------------------------------------------------------------------------
#
# Copyright (c) 2006-2020, ASSIMP Development Team
#
# All rights reserved.
#
# Redistribution and use of this software in source and binary forms, 
# with or without modification, are permitted provided that the following 
# conditions are met:
# 
# * Redistributions of source code must retain the above
#   copyright notice, this list of conditions and the
#   following disclaimer.
# 
# * Redistributions in binary form must reproduce the above
#   copyright notice, this list of conditions and the
#   following disclaimer in the documentation and/or other
#   materials provided with the distribution.
# 
# * Neither the name of the ASSIMP team, nor the names of its
#   contributors may be used to endorse or promote products
#   derived from this software without specific prior
#   written permission of the ASSIMP Development Team.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ---------------------------------------------------------------------------

"""Generate BlenderSceneGen.h and BlenderScene.cpp from the
data structures in BlenderScene.h to map from *any* DNA to
*our* DNA"""

import sys
import os
import re

inputfile      = os.path.join("..","..","code","BlenderScene.h")
outputfile_gen = os.path.join("..","..","code","BlenderSceneGen.h")
outputfile_src = os.path.join("..","..","code","BlenderScene.cpp")

template_gen = "BlenderSceneGen.h.template"
template_src = "BlenderScene.cpp.template"

# workaround for stackoverflowing when reading the linked list of scene objects
# with the usual approach. See embedded notes for details.
Structure_Convert_Base_fullcode = """
template <> void Structure::Convert<Base>( Base& dest, const FileDatabase& db ) const { 
	// note: as per https://github.com/assimp/assimp/issues/128,
	// reading the Object linked list recursively is prone to stack overflow.
	// This structure converter is therefore an hand-written exception that
	// does it iteratively.

	const int initial_pos = db.reader->GetCurrentPos();
	std::pair<Base*, int> todo = std::make_pair(&dest, initial_pos);
	Base* saved_prev = NULL;
	while(true) {
		Base& cur_dest = *todo.first;
		db.reader->SetCurrentPos(todo.second);

		// we know that this is a double-linked, circular list which we never
		// traverse backwards, so don't bother resolving the back links.
		cur_dest.prev = NULL;

		ReadFieldPtr<ErrorPolicy_Warn>(cur_dest.object,"*object",db);

		// just record the offset of the blob data and allocate storage.
		// Does _not_ invoke Convert() recursively.
		const int old = db.reader->GetCurrentPos();

		// the return value of ReadFieldPtr indicates whether the object 
		// was already cached. In this case, we don't need to resolve
		// it again.
		if(!ReadFieldPtr<ErrorPolicy_Warn>(cur_dest.next,"*next",db, true) && cur_dest.next) {
			todo = std::make_pair(&*cur_dest.next, db.reader->GetCurrentPos());
			continue;
		}
		break;
	}
	
	db.reader->SetCurrentPos(initial_pos + size);
}

"""


Structure_Convert_decl =  """
template <> void Structure :: Convert<{a}> (
    {a}& dest,
    const FileDatabase& db
    ) const
"""


Structure_Convert_ptrdecl = """
    ReadFieldPtr<{policy}>({destcast}dest.{name_canonical},"{name_dna}",db);"""

Structure_Convert_rawptrdecl = """
    {{
        boost::shared_ptr<{type}> {name_canonical};
        ReadFieldPtr<{policy}>({destcast}{name_canonical},"{name_dna}",db);
        dest.{name_canonical} = {name_canonical}.get();
    }}"""

Structure_Convert_arraydecl = """
    ReadFieldArray<{policy}>({destcast}dest.{name_canonical},"{name_dna}",db);"""

Structure_Convert_arraydecl2d = """
    ReadFieldArray2<{policy}>({destcast}dest.{name_canonical},"{name_dna}",db);"""

Structure_Convert_normal =  """
    ReadField<{policy}>({destcast}dest.{name_canonical},"{name_dna}",db);"""


DNA_RegisterConverters_decl = """
void DNA::RegisterConverters() """

DNA_RegisterConverters_add = """
    converters["{a}"] = DNA::FactoryPair( &Structure::Allocate<{a}>, &Structure::Convert<{a}> );"""


map_policy = {
     ""     : "ErrorPolicy_Igno"
    ,"IGNO" : "ErrorPolicy_Igno"
    ,"WARN" : "ErrorPolicy_Warn"
    ,"FAIL" : "ErrorPolicy_Fail"
}

#
def main():

    # -----------------------------------------------------------------------
    # Parse structure definitions from BlenderScene.h
    input = open(inputfile,"rt").read()

    #flags = re.ASCII|re.DOTALL|re.MULTILINE
    flags = re.DOTALL|re.MULTILINE
    #stripcoms = re.compile(r"/\*(.*?)*\/",flags)
    getstruct = re.compile(r"struct\s+(\w+?)\s*(:\s*ElemBase)?\s*\{(.*?)^\}\s*;",flags)
    getsmartx = re.compile(r"(std\s*::\s*)?(vector)\s*<\s*(boost\s*::\s*)?shared_(ptr)\s*<\s*(\w+)\s*>\s*>\s*",flags)
    getsmartp = re.compile(r"(boost\s*::\s*)?shared_(ptr)\s*<\s*(\w+)\s*>\s*",flags)
    getrawp   = re.compile(r"(\w+)\s*\*\s*",flags)
    getsmarta = re.compile(r"(std\s*::\s*)?(vector)\s*<\s*(\w+)\s*>\s*",flags)
    getpolicy = re.compile(r"\s*(WARN|FAIL|IGNO)",flags)
    stripenum = re.compile(r"enum\s+(\w+)\s*{.*?\}\s*;",flags)

    assert getsmartx and getsmartp and getsmarta and getrawp and getpolicy and stripenum
    
    enums = set()
    #re.sub(stripcoms," ",input)
    #print(input)

    hits = {}
    while 1:
        match = re.search(getstruct,input)
        if match is None:
            break

        tmp = match.groups()[2]
        while 1:
            match2 = re.search(stripenum,tmp)
            if match2 is None:
                break
            tmp = tmp[match2.end():]
            enums.add(match2.groups()[0])

        hits[match.groups()[0]] = list(
            filter(lambda x:x[:2] != "//" and len(x),
                map(str.strip,
                    re.sub(stripenum," ",match.groups()[2]).split(";")
        )))

        input = input[match.end():]

	for e in enums:
		print("Enum: "+e)
    for k,v in hits.items():
        out = []
        for line in v:
           
            policy = "IGNO"
            py = re.search(getpolicy,line) 
            if not py is None:
                policy = py.groups()[0]
                line = re.sub(getpolicy,"",line)

            ty = re.match(getsmartx,line) or re.match(getsmartp,line)  or\
                re.match(getsmarta,line) or re.match(getrawp,line)

            if ty is None:
                ty = line.split(None,1)[0]
            else:
                if len(ty.groups()) == 1:
                    ty = ty.groups()[-1] + "$" 
                elif ty.groups()[1] == "ptr":
                    ty = ty.groups()[2] + "*"
                elif ty.groups()[1] == "vector":
                    ty = ty.groups()[-1] + ("*" if len(ty.groups()) == 3 else "**")
                else:
                    assert False

            #print(line)
            sp = line.split(',')
            out.append((ty,sp[0].split(None)[-1].strip(),policy))
            for m in sp[1:]:
                out.append((ty,m.strip(),policy))
                
        v[:] = out
        print("Structure {0}".format(k))
        for elem in out:
			print("\t"+"\t".join(elem))
        print("")

   
    output = open(outputfile_gen,"wt")
    templt = open(template_gen,"rt").read()
    s = ""

    # -----------------------------------------------------------------------
    # Structure::Convert<T> declarations for all supported structures
    for k,v in hits.items():
        s += Structure_Convert_decl.format(a=k)+";\n";
    output.write(templt.replace("<HERE>",s))

    output = open(outputfile_src,"wt")
    templt = open(template_src,"rt").read()
    s = ""

    # -----------------------------------------------------------------------
    # Structure::Convert<T> definitions for all supported structures
    for k,v in hits.items():
    	s += "//" + "-"*80
    	if k == 'Base':
    		s += Structure_Convert_Base_fullcode 
    		continue
        s += Structure_Convert_decl.format(a=k)+ "{ \n";

        for type, name, policy in v:
            splits = name.split("[",1)
            name_canonical = splits[0]
            #array_part = "" if len(splits)==1 else "["+splits[1]
            is_raw_ptr = not not type.count("$")
            ptr_decl = "*"*(type.count("*") + (1 if is_raw_ptr else 0))
            
            name_dna = ptr_decl+name_canonical #+array_part

            #required  = "false"
            policy = map_policy[policy]
            destcast = "(int&)" if type in enums else ""

            # POINTER
            if is_raw_ptr:
                type = type.replace('$','')
                s += Structure_Convert_rawptrdecl.format(**locals())
            elif ptr_decl:
                s += Structure_Convert_ptrdecl.format(**locals())
            # ARRAY MEMBER
            elif name.count('[')==1:
                s += Structure_Convert_arraydecl.format(**locals())
            elif name.count('[')==2:
                s += Structure_Convert_arraydecl2d.format(**locals())
            # NORMAL MEMBER
            else:
                s += Structure_Convert_normal.format(**locals())

        s += "\n\n\tdb.reader->IncPtr(size);\n}\n\n"


    # -----------------------------------------------------------------------
    # DNA::RegisterConverters - collect all available converter functions
    # in a std::map<name,converter_proc>
    #s += "#if 0\n"
    s += "//" + "-"*80 + DNA_RegisterConverters_decl + "{\n"
    for k,v in hits.items():
        s += DNA_RegisterConverters_add.format(a=k)
        
    s += "\n}\n"
    #s += "#endif\n"
        
    output.write(templt.replace("<HERE>",s))

    # we got here, so no error
    return 0

if __name__ == "__main__":
    sys.exit(main())



from gibson.core.physics.robot_locomotors import Turtlebot
from gibson.envs.mobile_robots_env import TurtlebotNavigateEnv
from gibson.envs.mobile_robots_env import tracking_camera


class TurtlebotNavigateNoPhysicsEnv(TurtlebotNavigateEnv):
    def __init__(self, config, gpu_idx=0):
        TurtlebotNavigateEnv.__init__(self, config, gravity=0.0, collision_enabled=False)



import torch
import math
import random
from PIL import Image, ImageOps

try:
    import accimage
except ImportError:
    accimage = None
import numpy as np
import numbers
import types
import collections


class RandomScale(object):
    """Rescale the input PIL.Image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (w, h), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, minsize, maxsize, interpolation=Image.BILINEAR):
        assert isinstance(minsize, int)
        assert isinstance(maxsize, int)
        self.minsize = minsize
        self.maxsize = maxsize
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Args:
            img (PIL.Image): Image to be scaled.
        Returns:
            PIL.Image: Rescaled image.
        """

        size = random.randint(self.minsize, self.maxsize)

        if isinstance(size, int):
            w, h = img.size
            if (w <= h and w == size) or (h <= w and h == size):
                return img
            if w < h:
                ow = size
                oh = int(size * h / w)
                return img.resize((ow, oh), self.interpolation)
            else:
                oh = size
                ow = int(size * w / h)
                return img.resize((ow, oh), self.interpolation)
        else:
            raise NotImplementedError()



import numpy as np
import tensorflow as tf
from baselines.a2c.utils import conv, fc, conv_to_fc, batch_to_seq, seq_to_batch, lstm, lnlstm
from baselines.common.distributions import make_pdtype
import gym.spaces
import baselines.common.tf_util as U


## Fuse policy using PPO2 from OpenAI Baseline

class FusePolicy(object):
    def __init__(self, sess, ob_space, sensor_space, ac_space, nbatch, nsteps, reuse=False): #pylint: disable=W0613
        if isinstance(ac_space, gym.spaces.Discrete):
            self.is_discrete = True
        else:
            self.is_discrete = False

        ob_shape = (nbatch,) + ob_space.shape
        ob_sensor_shape = (nbatch,) + sensor_space.shape
        if self.is_discrete:
            actdim = ac_space.n
        else:
            actdim =  ac_space.shape[0]
        X_camera = tf.placeholder(tf.uint8, ob_shape, name='Ob_camera') #obs
        X_sensor = tf.placeholder(tf.float32, ob_sensor_shape, name='Ob_sensor')

        self.pdtype = make_pdtype(ac_space)

        with tf.variable_scope("model", reuse=reuse):
            h_camera = conv(tf.cast(X_camera, tf.float32)/255., 'c1', nf=32, rf=8, stride=4, init_scale=np.sqrt(2))
            h2_camera = conv(h_camera, 'c2', nf=64, rf=4, stride=2, init_scale=np.sqrt(2))
            h3_camera = conv(h2_camera, 'c3', nf=64, rf=3, stride=1, init_scale=np.sqrt(2))
            h3_camera = conv_to_fc(h3_camera)
            h4_camera = fc(h3_camera, 'fc1', nh=512, init_scale=np.sqrt(2))
            pi_camera = fc(h4_camera, 'pi', actdim, init_scale=0.01)
            vf_camera = fc(h4_camera, 'v', 1)[:,0]

        self.pd = self.pdtype.pdfromflat(pi_camera)

        with tf.variable_scope("model_sensor", reuse=reuse):
            h1_sensor = fc(X_sensor, 'pi_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2_sensor = fc(h1_sensor, 'pi_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            pi_sensor = fc(h2_sensor, 'pi', actdim, init_scale=0.01)
            h1_sensor = fc(X_sensor, 'vf_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2_sensor = fc(h1_sensor, 'vf_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            vf_sensor = fc(h2_sensor, 'vf', 1)[:,0]

        with tf.variable_scope("model", reuse=reuse):
            logstd = tf.get_variable(name="logstd", shape=[1, actdim], 
                initializer=tf.zeros_initializer())
            X = tf.concat([X_camera, X_sensor], 0)
            pi_full = tf.concat([pi_camera, pi_sensor], 0)
            pi = fc(pi_full, 'pi', actdim, init_scale=0.01)
            vf_full = tf.concat([vf_camera, vf_sensor], 0)
            vf = fc(vf_full, 'vf', 1)[:,0]

        pdparam = tf.concat([pi, pi * 0.0 + logstd], axis=1)
        self.pd = self.pdtype.pdfromflat(pdparam)

        a0 = self.pd.sample()
        neglogp0 = self.pd.neglogp(a0)
        self.initial_state = None

        def step(ob, ob_sensor, *_args, **_kwargs):
            a, v, neglogp = sess.run([a0, vf, neglogp0], {X_camera:ob, X_sensor: ob_sensor})
            return a, v, self.initial_state, neglogp

        def value(ob, ob_sensor, *_args, **_kwargs):
            return sess.run(vf, {X_camera:ob, X_sensor: ob_sensor})

        self.X = X
        self.pi = pi
        self.vf = vf
        self.step = step
        self.value = value



class CnnPolicy(object):

    def __init__(self, sess, ob_space, ac_space, nbatch, nsteps, reuse=False, is_discrete=False): #pylint: disable=W0613
        if isinstance(ac_space, gym.spaces.Discrete):
            self.is_discrete = True
        else:
            self.is_discrete = False

        print("nbatch%d" % (nbatch))

        nh, nw, nc = ob_space.shape
        ob_shape = (nbatch, nh, nw, nc)
        if self.is_discrete:
            nact = ac_space.n
        else:
            nact =  ac_space.shape[0]
        X = tf.placeholder(tf.float32, ob_shape) #obs
        
        with tf.variable_scope("model", reuse=reuse):
            h = conv(X, 'c1', nf=32, rf=8, stride=4, init_scale=np.sqrt(2))
            h2 = conv(h, 'c2', nf=64, rf=4, stride=2, init_scale=np.sqrt(2))
            h3 = conv(h2, 'c3', nf=64, rf=3, stride=1, init_scale=np.sqrt(2))
            h3 = conv_to_fc(h3)
            h4 = fc(h3, 'fc1', nh=512, init_scale=np.sqrt(2))
            h5 = fc(h3, 'fc_vf', nh=512, init_scale=np.sqrt(2))


            pi = fc(h4, 'pi', nact, init_scale=0.05)
            vf = fc(h5, 'v', 1, act=lambda x: x)[:,0]

            if not self.is_discrete:
                logstd = tf.get_variable(name="logstd", shape=[1, nact],
                                     initializer=tf.zeros_initializer())

        self.pdtype = make_pdtype(ac_space)
        if self.is_discrete:
            self.pd = self.pdtype.pdfromflat(pi)
            a0 = self.pd.sample()
        else:
            pdparam = tf.concat([pi, pi * 0.0 + logstd], axis=1)
            self.pd = self.pdtype.pdfromflat(pdparam)
            a0 = self.pd.sample()

        neglogp0 = self.pd.neglogp(a0)
        self.initial_state = None

        def step(ob, *_args, **_kwargs):
            a, v, neglogp = sess.run([a0, vf, neglogp0], {X:ob.astype(np.float32)/255.0})
            assert(a.shape[0] == 1) # make sure a = a[0] don't throw away actions
            a = a[0]
            print(a,v, neglogp)
            return a, v, self.initial_state, neglogp

        def value(ob, *_args, **_kwargs):
            return sess.run(vf, {X:ob})


        self.X = X
        self.pi = pi
        self.vf = vf
        self.step = step
        self.value = value


class MlpPolicy(object):
    def __init__(self, sess, ob_space, ac_space, nbatch, nsteps, reuse=False): #pylint: disable=W0613
        if isinstance(ac_space, gym.spaces.Discrete):
            self.is_discrete = True
        else:
            self.is_discrete = False

        ob_shape = (nbatch,) + ob_space.shape
        if self.is_discrete:
            actdim = ac_space.n
        else:
            actdim =  ac_space.shape[0]
        X = tf.placeholder(tf.float32, ob_shape, name='Ob') #obs
        with tf.variable_scope("model", reuse=reuse):
            h1 = fc(X, 'pi_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2 = fc(h1, 'pi_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            pi = fc(h2, 'pi', actdim, init_scale=0.01)
            h1 = fc(X, 'vf_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2 = fc(h1, 'vf_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            vf = fc(h2, 'vf', 1)[:,0]
            logstd = tf.get_variable(name="logstd", shape=[1, actdim], 
                initializer=tf.zeros_initializer())

        pdparam = tf.concat([pi, pi * 0.0 + logstd], axis=1)

        self.pdtype = make_pdtype(ac_space)
        if self.is_discrete:
            self.pd = self.pdtype.pdfromflat(pi)
            a0 = self.pd.sample()
        else:
            self.pd = self.pdtype.pdfromflat(pdparam)
            a0 = self.pd.sample()

        neglogp0 = self.pd.neglogp(a0)
        self.initial_state = None

        def step(ob, *_args, **_kwargs):
            a, v, neglogp = sess.run([a0, vf, neglogp0], {X:ob})
            a = a[0]
            return a, v, self.initial_state, neglogp

        def value(ob, *_args, **_kwargs):
            return sess.run(vf, {X:ob})

        self.X = X
        self.pi = pi
        self.vf = vf
        self.step = step
        self.value = value



class MlpPolicy2(object):
    def __init__(self, sess, ob_space, ac_space, nbatch, nsteps, reuse=False): #pylint: disable=W0613
        if isinstance(ac_space, gym.spaces.Discrete):
            self.is_discrete = True
        else:
            self.is_discrete = False

        ob_shape = (nbatch,) + ob_space.shape
        if self.is_discrete:
            actdim = ac_space.n
        else:
            actdim =  ac_space.shape[0]
        X = tf.placeholder(tf.float32, ob_shape, name='Ob') #obs
        with tf.variable_scope("model", reuse=reuse):
            X2 = conv_to_fc(X)
            h1 = fc(X2, 'pi_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2 = fc(h1, 'pi_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            pi = fc(h2, 'pi', actdim, init_scale=0.01)
            h1 = fc(X2, 'vf_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2 = fc(h1, 'vf_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            vf = fc(h2, 'vf', 1)[:,0]
            logstd = tf.get_variable(name="logstd", shape=[1, actdim], 
                initializer=tf.zeros_initializer())

        pdparam = tf.concat([pi, pi * 0.0 + logstd], axis=1)

        self.pdtype = make_pdtype(ac_space)
        if self.is_discrete:
            self.pd = self.pdtype.pdfromflat(pi)
            a0 = self.pd.sample()
        else:
            self.pd = self.pdtype.pdfromflat(pdparam)
            a0 = self.pd.sample()

        neglogp0 = self.pd.neglogp(a0)
        self.initial_state = None

        def step(ob, *_args, **_kwargs):
            a, v, neglogp = sess.run([a0, vf, neglogp0], {X:ob.astype(np.float32) / 255.0})
            a = a[0]
            return a, v, self.initial_state, neglogp

        def value(ob, *_args, **_kwargs):
            return sess.run(vf, {X:ob})

        self.X = X
        self.pi = pi
        self.vf = vf
        self.step = step
        self.value = value


class CnnPolicy2(object):
    def __init__(self, sess, ob_space, ac_space, nbatch, nsteps, reuse=False): #pylint: disable=W0613
        if isinstance(ac_space, gym.spaces.Discrete):
            self.is_discrete = True
        else:
            self.is_discrete = False

        ob_shape = (nbatch,) + ob_space.shape
        if self.is_discrete:
            actdim = ac_space.n
        else:
            actdim =  ac_space.shape[0]
        
        X = tf.placeholder(tf.float32, ob_shape, name='Ob') #obs
        with tf.variable_scope("model", reuse=reuse):
            h_c = conv(X, 'c1', nf=32, rf=8, stride=4, init_scale=np.sqrt(2))
            h2_c = conv(h_c, 'c2', nf=64, rf=4, stride=2, init_scale=np.sqrt(2))
            h2_c = conv_to_fc(h_c)
            h1 = fc(h2_c, 'pi_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2 = fc(h1, 'pi_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            pi = fc(h2, 'pi', actdim, init_scale=0.01)
            h1 = fc(h2_c, 'vf_fc1', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            h2 = fc(h1, 'vf_fc2', nh=64, init_scale=np.sqrt(2), act=tf.tanh)
            vf = fc(h2, 'vf', 1)[:,0]
            logstd = tf.get_variable(name="logstd", shape=[1, actdim], 
                initializer=tf.zeros_initializer())

        pdparam = tf.concat([pi, pi * 0.0 + logstd], axis=1)

        self.pdtype = make_pdtype(ac_space)
        if self.is_discrete:
            self.pd = self.pdtype.pdfromflat(pi)
            a0 = self.pd.sample()
        else:
            self.pd = self.pdtype.pdfromflat(pdparam)
            a0 = self.pd.sample()

        neglogp0 = self.pd.neglogp(a0)
        self.initial_state = None

        def step(ob, *_args, **_kwargs):
            a, v, neglogp = sess.run([a0, vf, neglogp0], {X:ob.astype(np.float32) / 255.0})
            a = a[0]
            return a, v, self.initial_state, neglogp

        def value(ob, *_args, **_kwargs):
            return sess.run(vf, {X:ob})

        self.X = X
        self.pi = pi
        self.vf = vf
        self.step = step
        self.value = value




## Issue related to time resolution/smoothness
#  http://bulletphysics.org/mediawiki-1.5.8/index.php/Stepping_The_World

from gibson.core.physics.scene_building import SinglePlayerBuildingScene
from gibson.core.physics.scene_stadium import SinglePlayerStadiumScene
import pybullet as p
import time
import random
import zmq
import math
import argparse
import os
import json
import numpy as np
from transforms3d import euler, quaternions
from gibson.core.physics.physics_object import PhysicsObject
from gibson.core.render.profiler import Profiler
import gym, gym.spaces, gym.utils, gym.utils.seeding
import sys
import yaml

class BaseEnv(gym.Env):
    """
    Base class for loading environments in a Scene.
    Handles scene loading, starting physical simulation

    These environments create single-player scenes and behave like normal Gym environments.
    Multiplayer is not yet supported
    """

    def __init__(self, config, scene_type, tracking_camera):
        ## Properties already instantiated from SensorEnv/CameraEnv
        #   @self.robot
        self.gui = config["mode"] == "gui"
        self.model_id = config["model_id"]
        self.timestep = config["speed"]["timestep"]
        self.frame_skip = config["speed"]["frameskip"]
        self.resolution = config["resolution"]
        self.tracking_camera = tracking_camera
        self.robot = None
        target_orn, target_pos   = config["target_orn"], self.config["target_pos"]
        initial_orn, initial_pos = config["initial_orn"], self.config["initial_pos"]

        if config["display_ui"]:
            #self.physicsClientId = p.connect(p.DIRECT)
            self.physicsClientId = p.connect(p.GUI, "--opengl2")
            p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
        elif (self.gui):
            self.physicsClientId = p.connect(p.GUI, "--opengl2")
        else:
            self.physicsClientId = p.connect(p.DIRECT)

        self.camera = Camera()
        self._seed()
        self._cam_dist = 3
        self._cam_yaw = 0
        self._cam_pitch = -30
        self.scene_type = scene_type
        self.scene = None
    
    def _close(self):
        p.disconnect()

    def parse_config(self, config):
        with open(config, 'r') as f:
            config_data = yaml.load(f)
        return config_data
        
    def create_scene(self, gravity=9.8, collision_enabled=True):
        if self.scene is not None:
            return
        if self.scene_type == "stadium":
            self.scene = self.create_single_player_stadium_scene()
        elif self.scene_type == "building":
            self.scene = self.create_single_player_building_scene(gravity=gravity, collision_enabled=collision_enabled)
        else:
            raise AssertionError()
        self.robot.scene = self.scene
    
    def create_single_player_building_scene(self, gravity=9.8, collision_enabled=True):
        return SinglePlayerBuildingScene(self.robot, model_id=self.model_id, gravity=gravity, timestep=self.timestep, frame_skip=self.frame_skip, collision_enabled=collision_enabled, env=self)

    def create_single_player_stadium_scene(self):
        return SinglePlayerStadiumScene(self.robot, gravity=9.8, timestep=self.timestep, frame_skip=self.frame_skip, env=self)


    def configure(self, args):
        self.robot.args = args
    
    def _seed(self, seed=None):
        self.np_random, seed = gym.utils.seeding.np_random(seed)
        return [seed]

    def _reset(self):
        assert self.robot is not None, "Pleases introduce robot to environment before resetting."
        p.configureDebugVisualizer(p.COV_ENABLE_GUI,0)
        p.configureDebugVisualizer(p.COV_ENABLE_KEYBOARD_SHORTCUTS, 0)
        p.configureDebugVisualizer(p.COV_ENABLE_MOUSE_PICKING, 1)
        p.configureDebugVisualizer(p.COV_ENABLE_SHADOWS, 1)
        p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1)

        self.frame = 0
        self.done = 0
        self.reward = 0
        dump = 0
        state = self.robot.reset()
        self.scene.episode_restart()
        return state

    def _render(self, mode, close):
        base_pos=[0,0,0]
        if (hasattr(self,'robot')):
            if (hasattr(self.robot,'body_xyz')):
                base_pos = self.robot.body_xyz
        
        view_matrix = p.computeViewMatrixFromYawPitchRoll(
            cameraTargetPosition=base_pos,
            distance=self._cam_dist,
            yaw=self._cam_yaw,
            pitch=self._cam_pitch,
            roll=0,
            upAxisIndex=2)
        proj_matrix = p.computeProjectionMatrixFOV(
            fov=60, aspect=float(self._render_width)/self._render_height,
            nearVal=0.1, farVal=100.0)
        (_, _, px, _, _) = p.getCameraImage(
        width=self._render_width, height=self._render_height, viewMatrix=view_matrix,
            projectionMatrix=proj_matrix,
            renderer=p.ER_BULLET_HARDWARE_OPENGL
            )
        rgb_array = np.array(px).reshape((self._render_width, self._render_height, -1))
        if close: return None
        rgb_array = rgb_array[:, :, :3]
        return rgb_array

    def render_physics(self):
        robot_pos, _ = p.getBasePositionAndOrientation(self.robot_tracking_id)
        
        view_matrix = p.computeViewMatrixFromYawPitchRoll(
            cameraTargetPosition=robot_pos,
            distance=self.tracking_camera["distance"],
            yaw=self.tracking_camera["yaw"],
            pitch=self.tracking_camera["pitch"],
            roll=0,
            upAxisIndex=2)
        proj_matrix = p.computeProjectionMatrixFOV(
            fov=60, aspect=float(self._render_width)/self._render_height,
            nearVal=0.1, farVal=100.0)
        with Profiler("render physics: Get camera image"):
            (_, _, px, _, _) = p.getCameraImage(
            width=self._render_width, height=self._render_height, viewMatrix=view_matrix,
                projectionMatrix=proj_matrix,
                renderer=p.ER_TINY_RENDERER
                )
        rgb_array = np.array(px).reshape((self._render_width, self._render_height, -1))
        rgb_array = rgb_array[:, :, :3]
        return rgb_array


    def render_map(self):
        base_pos=[0, 0, -3]
        if (hasattr(self,'robot')):
            if (hasattr(self.robot,'body_xyz')):
                base_pos[0] = self.robot.body_xyz[0]
                base_pos[1] = self.robot.body_xyz[1]
        
        view_matrix = p.computeViewMatrixFromYawPitchRoll(
            cameraTargetPosition=base_pos,
            distance=35,
            yaw=0,
            pitch=-89,
            roll=0,
            upAxisIndex=2)
        proj_matrix = p.computeProjectionMatrixFOV(
            fov=60, aspect=float(self._render_width)/self._render_height,
            nearVal=0.1, farVal=100.0)
        (_, _, px, _, _) = p.getCameraImage(
        width=self._render_width, height=self._render_height, viewMatrix=view_matrix,
            projectionMatrix=proj_matrix,
            renderer=p.ER_BULLET_HARDWARE_OPENGL
            )
        rgb_array = np.array(px).reshape((self._render_width, self._render_height, -1))
        rgb_array = rgb_array[:, :, :3]
        return rgb_array

    def get_action_dim(self):
        return len(self.robot.ordered_joints)

    def get_observation_dim(self):
        return 1

    def _close(self):
        if (self.physicsClientId>=0):
            p.disconnect(self.physicsClientId)
            self.physicsClientId = -1
    
    def set_window(self, posX, posY, sizeX, sizeY):
        values = {      
            'name': "Robot",  
            'gravity': 0,
            'posX': int(posX),
            'posY': int(posY),
            'sizeX': int(sizeX),
            'sizeY': int(sizeY)
        }
        cmd = 'wmctrl -r \"Bullet Physics\" -e {gravity},{posX},{posY},{sizeX},{sizeY}'.format(**values)
        os.system(cmd)

        cmd = "xdotool search --name \"Bullet Physics\" set_window --name \"Robot's world\""
        os.system(cmd)


class Camera:
    def __init__(self):
        pass

    def move_and_look_at(self,i,j,k,x,y,z):
        lookat = [x,y,z]
        distance = 10
        yaw = 10




"""This file implements the functionalities of a minitaur using pybullet.

"""
import copy
import math
import numpy as np
from gibson.core.physics import motor
from gibson.core.physics.robot_locomotors import WalkerBase
from gibson.core.physics.robot_bases import Joint, BodyPart
import os, sys
import pybullet as p
import gym
from transforms3d.euler import euler2quat


tracking_camera = {
    'yaw': 20,
    'z_offset': 0.3,
    'distance': 2,
    'pitch': -20
}


class MinitaurBase(WalkerBase):
    model_type = "URDF"
    default_scale = 1
    
    KNEE_CONSTRAINT_POINT_RIGHT = [0, 0.005, 0.2]
    KNEE_CONSTRAINT_POINT_LEFT = [0, 0.01, 0.2]
    OVERHEAT_SHUTDOWN_TORQUE = 2.45
    OVERHEAT_SHUTDOWN_TIME = 1.0
    LEG_POSITION = ["front_left", "back_left", "front_right", "back_right"]
    MOTOR_NAMES = [
        "motor_front_leftL_joint", "motor_front_leftR_joint",
        "motor_back_leftL_joint", "motor_back_leftR_joint",
        "motor_front_rightL_joint", "motor_front_rightR_joint",
        "motor_back_rightL_joint", "motor_back_rightR_joint"
    ]
    LEG_LINK_ID = [2, 3, 5, 6, 8, 9, 11, 12, 15, 16, 18, 19, 21, 22, 24, 25]
    MOTOR_LINK_ID = [1, 4, 7, 10, 14, 17, 20, 23]
    FOOT_LINK_ID = [3, 6, 9, 12, 16, 19, 22, 25]
    BASE_LINK_ID = -1
    OBSERVATION_DIM = 3 * len(MOTOR_NAMES) + 4   # VELOCITY, ANGLE, TORQUES

    self_collision_enabled=True
    motor_velocity_limit=np.inf
    
    #accurate_motor_model_enabled=False   ## (hzyjerry): affect speed?
    motor_kp=1.00
    motor_kd=0.2
    torque_control_enabled=False
    motor_overheat_protection=True
    on_rack=False
    kd_for_pd_controllers=0.3 
    mjcf_scaling = 1
    num_motors = 8
    num_legs = int(num_motors / 2)
    motor_direction = [-1, -1, -1, -1, 1, 1, 1, 1]
    observed_motor_torques = np.zeros(num_motors)
    applied_motor_torques = np.zeros(num_motors)
    max_force = 5.5
    joint_name_to_id = None
        
    """The minitaur class that simulates a quadruped robot from Ghost Robotics.
    """

    def __init__(self, config, env=None,
                 pd_control_enabled=True,
                 accurate_motor_model_enabled=True):
        """Constructs a minitaur and reset it to the initial states.

        Properties:
        self_collision_enabled: Whether to enable self collision.
        motor_velocity_limit: The upper limit of the motor velocity.
        pd_control_enabled: Whether to use PD control for the motors. If true, need smaller time step to stablize (1/500.0 timestep)
        accurate_motor_model_enabled: Whether to use the accurate DC motor model.
        motor_kp: proportional gain for the accurate motor model
        motor_kd: derivative gain for the acurate motor model
        torque_control_enabled: Whether to use the torque control, if set to
            False, pose control will be used.
        motor_overheat_protection: Whether to shutdown the motor that has exerted
            large torque (OVERHEAT_SHUTDOWN_TORQUE) for an extended amount of time
            (OVERHEAT_SHUTDOWN_TIME). See apply_action() in minitaur.py for more
            details.
        on_rack: Whether to place the minitaur on rack. This is only used to debug
            the walking gait. In this mode, the minitaur's base is hanged midair so
            that its walking gait is clearer to visualize.
        kd_for_pd_controllers: kd value for the pd controllers of the motors.
        """
        self.config = config
        self.model_type = "URDF"
        #self.robot_name = "quadruped"
        self.robot_name = "base_chassis_link"
        scale = config["robot_scale"] if "robot_scale" in config.keys() else self.default_scale
        
        WalkerBase.__init__(self, 
                            "quadruped/minitaur.urdf", 
                            self.robot_name, 
                            action_dim=8, 
                            sensor_dim=self.OBSERVATION_DIM, 
                            power=5,
                            scale = scale,
                            initial_pos=config['initial_pos'],
                            target_pos=config["target_pos"], 
                            resolution=config["resolution"], 
                            env = env)
        
        self.r_f = 0.1
        self.time_step = config["speed"]["timestep"]
        self.pd_control_enabled = pd_control_enabled
        self.minitaur = None   ## TODO: fix this
        self.accurate_motor_model_enabled = accurate_motor_model_enabled
        if self.accurate_motor_model_enabled:
            self._kp = self.motor_kp
            self._kd = self.motor_kd
            self._motor_model = motor.MotorModel(torque_control_enabled=self.torque_control_enabled,kp=self._kp,kd=self._kd)
        elif self.pd_control_enabled:
            self._kp = 8
            self._kd = self.kd_for_pd_controllers
        else:
            self._kp = 1
            self._kd = 1

        if config["is_discrete"]:
            self.action_space = gym.spaces.Discrete(17)
            self.torque = 10
            ## Hip_1, Ankle_1, Hip_2, Ankle_2, Hip_3, Ankle_3, Hip_4, Ankle_4 
            self.action_list = [[self.r_f * self.torque, 0, 0, 0, 0, 0, 0, 0],
                                [0, self.r_f * self.torque, 0, 0, 0, 0, 0, 0],
                                [0, 0, self.r_f * self.torque, 0, 0, 0, 0, 0],
                                [0, 0, 0, self.r_f * self.torque, 0, 0, 0, 0],
                                [0, 0, 0, 0, self.r_f * self.torque, 0, 0, 0],
                                [0, 0, 0, 0, 0, self.r_f * self.torque, 0, 0],
                                [0, 0, 0, 0, 0, 0, self.r_f * self.torque, 0],
                                [0, 0, 0, 0, 0, 0, 0, self.r_f * self.torque],
                                [-self.r_f * self.torque, 0, 0, 0, 0, 0, 0, 0],
                                [0, -self.r_f * self.torque, 0, 0, 0, 0, 0, 0],
                                [0, 0, -self.r_f * self.torque, 0, 0, 0, 0, 0],
                                [0, 0, 0, -self.r_f * self.torque, 0, 0, 0, 0],
                                [0, 0, 0, 0, -self.r_f * self.torque, 0, 0, 0],
                                [0, 0, 0, 0, 0, -self.r_f * self.torque, 0, 0],
                                [0, 0, 0, 0, 0, 0, -self.r_f * self.torque, 0],
                                [0, 0, 0, 0, 0, 0, 0, -self.r_f * self.torque],
                                [0, 0, 0, 0, 0, 0, 0, 0]]
            self.setup_keys_to_action()
        self.debug_count = 0
        self.qmax = [0] * 8
        self.fmax = [0] * 8
    

    def _RecordMassInfoFromURDF(self):
        self._base_mass_urdf = p.getDynamicsInfo(self.minitaur, self.BASE_LINK_ID)[0]
        self._leg_masses_urdf = []
        self._leg_masses_urdf.append(p.getDynamicsInfo(self.minitaur, self.LEG_LINK_ID[0])[0])
        self._leg_masses_urdf.append(p.getDynamicsInfo(self.minitaur, self.MOTOR_LINK_ID[0])[0])

    def _BuildJointNameToIdDict(self):
        num_joints = p.getNumJoints(self.minitaur)
        self.joint_name_to_id = {}
        for i in range(num_joints):
            joint_info = p.getJointInfo(self.minitaur, i)
            self.joint_name_to_id[joint_info[1].decode("UTF-8")] = joint_info[0]

    def _BuildMotorIdList(self):
        self._motor_id_list = [
            self.joint_name_to_id[motor_name] for motor_name in self.MOTOR_NAMES
        ]


    def robot_specific_reset(self, reload_urdf=True):
        """Reset the minitaur to its initial states.

        Args:
          reload_urdf: Whether to reload the urdf file. If not, Reset() just place
            the minitaur back to its starting position.
        """
        if self.minitaur is None:
            self.minitaur = self.robot_ids[0]

        if self.joint_name_to_id is None:
            self._BuildJointNameToIdDict()
            self._BuildMotorIdList()
            self._RecordMassInfoFromURDF()
            self.ResetPose(add_constraint=True)
            self._overheat_counter = np.zeros(self.num_motors)
            self._motor_enabled_list = [True] * self.num_motors
            if self.on_rack:
                p.createConstraint(self.minitaur, -1, -1, -1, p.JOINT_FIXED,[0, 0, 0], [0, 0, 0], [0, 0, 1])
        self.ResetPose(add_constraint=True)

    def _SetMotorTorqueById(self, motor_id, torque):
        p.setJointMotorControl2(bodyIndex=self.minitaur,
                                jointIndex=motor_id,
                                controlMode=p.TORQUE_CONTROL,
                                force=torque)

    def _SetDesiredMotorAngleById(self, motor_id, desired_angle):
        p.setJointMotorControl2(bodyIndex=self.minitaur,
                                jointIndex=motor_id,
                                controlMode=p.POSITION_CONTROL,
                                targetPosition=desired_angle,
                                positionGain=self._kp,
                                velocityGain=self._kd,
                                force=self.max_force)

    def _SetDesiredMotorAngleByName(self, motor_name, desired_angle):
        self._SetDesiredMotorAngleById(self.joint_name_to_id[motor_name],desired_angle)

    def calc_potential(self):
        return 0


    def setup_keys_to_action(self):
        self.keys_to_action = {
            (ord('s'), ): 0, ## backward
            (ord('w'), ): 1, ## forward
            (ord('d'), ): 2, ## turn right
            (ord('a'), ): 3, ## turn left
            (): 4
        }

    def ResetPose(self, add_constraint):
        """Reset the pose of the minitaur.

        Args:
          add_constraint: Whether to add a constraint at the joints of two feet.
        """
        for i in range(self.num_legs):
            self._ResetPoseForLeg(i, add_constraint)

    def _ResetPoseForLeg(self, leg_id, add_constraint):
        """Reset the initial pose for the leg.

        Args:
          leg_id: It should be 0, 1, 2, or 3, which represents the leg at
            front_left, back_left, front_right and back_right.
          add_constraint: Whether to add a constraint at the joints of two feet.
        """
        knee_friction_force = 0
        half_pi = math.pi / 2.0
        knee_angle = -2.1834

        leg_position = self.LEG_POSITION[leg_id]
        p.resetJointState(
            self.minitaur,
            self.joint_name_to_id["motor_" + leg_position + "L_joint"],
            self.motor_direction[2 * leg_id] * half_pi,
            targetVelocity=0)
        p.resetJointState(
            self.minitaur,
            self.joint_name_to_id["knee_" + leg_position + "L_link"],
            self.motor_direction[2 * leg_id] * knee_angle,
            targetVelocity=0)
        p.resetJointState(
            self.minitaur,
            self.joint_name_to_id["motor_" + leg_position + "R_joint"],
            self.motor_direction[2 * leg_id + 1] * half_pi,
            targetVelocity=0)
        p.resetJointState(
            self.minitaur,
            self.joint_name_to_id["knee_" + leg_position + "R_link"],
            self.motor_direction[2 * leg_id + 1] * knee_angle,
            targetVelocity=0)
        if add_constraint:
            p.createConstraint(self.minitaur, 
                               self.joint_name_to_id["knee_" + leg_position + "R_link"],
                               self.minitaur, 
                               self.joint_name_to_id["knee_" + leg_position + "L_link"],
                               p.JOINT_POINT2POINT, 
                               [0, 0, 0],
                               self.KNEE_CONSTRAINT_POINT_RIGHT, 
                               self.KNEE_CONSTRAINT_POINT_LEFT)

        if self.accurate_motor_model_enabled or self.pd_control_enabled:
            # Disable the default motor in pybullet.
            p.setJointMotorControl2(bodyIndex=self.minitaur,
                                    jointIndex=(self.joint_name_to_id["motor_" + leg_position + "L_joint"]),
                                    controlMode=p.VELOCITY_CONTROL,
                                    targetVelocity=0,
                                    force=knee_friction_force)
            p.setJointMotorControl2(bodyIndex=self.minitaur,
                                    jointIndex=(self.joint_name_to_id["motor_" + leg_position + "R_joint"]),
                                    controlMode=p.VELOCITY_CONTROL,
                                    targetVelocity=0,
                                    force=knee_friction_force)

        else:
            self._SetDesiredMotorAngleByName("motor_" + leg_position + "L_joint",
                                             self.motor_direction[2 * leg_id] * half_pi)
            self._SetDesiredMotorAngleByName("motor_" + leg_position + "R_joint",
                                             self.motor_direction[2 * leg_id + 1] * half_pi)

        p.setJointMotorControl2(bodyIndex=self.minitaur,
                                jointIndex=(self.joint_name_to_id["knee_" + leg_position + "L_link"]),
                                controlMode=p.VELOCITY_CONTROL,
                                targetVelocity=0,
                                force=knee_friction_force)
        p.setJointMotorControl2(bodyIndex=self.minitaur,
                                jointIndex=(self.joint_name_to_id["knee_" + leg_position + "R_link"]),
                                controlMode=p.VELOCITY_CONTROL,
                                targetVelocity=0,
                                force=knee_friction_force)

    def GetBasePosition(self):
        """Get the position of minitaur's base.

        Returns:
          The position of minitaur's base.
        """
        position, _ = (
            p.getBasePositionAndOrientation(self.minitaur))
        return position

    def GetBaseOrientation(self):
        """Get the orientation of minitaur's base, represented as quaternion.

        Returns:
          The orientation of minitaur's base.
        """
        _, orientation = (
            p.getBasePositionAndOrientation(self.minitaur))
        return orientation

    def GetActionDimension(self):
        """Get the length of the action list.

        Returns:
          The length of the action list.
        """
        return self.num_motors

    def GetObservationUpperBound(self):
        """Get the upper bound of the observation.

        Returns:
          The upper bound of an observation. See GetObservation() for the details
            of each element of an observation.
        """
        upper_bound = np.array([0.0] * self.GetObservationDimension())
        upper_bound[0:self.num_motors] = math.pi  # Joint angle.
        upper_bound[self.num_motors:2 * self.num_motors] = (
            motor.MOTOR_SPEED_LIMIT)  # Joint velocity.
        upper_bound[2 * self.num_motors:3 * self.num_motors] = (
            motor.OBSERVED_TORQUE_LIMIT)  # Joint torque.
        upper_bound[3 * self.num_motors:] = 1.0  # Quaternion of base orientation.
        return upper_bound

    def GetObservationLowerBound(self):
        """Get the lower bound of the observation."""
        return -self.GetObservationUpperBound()

    def GetObservationDimension(self):
        """Get the length of the observation list.

        Returns:
          The length of the observation list.
        """
        return len(self.GetObservation())

    def calc_state(self):
        return self.GetObservation()

    def GetObservation(self):
        """Get the observations of minitaur.

        It includes the angles, velocities, torques and the orientation of the base.

        Returns:
          The observation list. observation[0:8] are motor angles. observation[8:16]
          are motor velocities, observation[16:24] are motor torques.
          observation[24:28] is the orientation of the base, in quaternion form.
        """
        observation = []
        observation.extend(self.GetMotorAngles().tolist())
        observation.extend(self.GetMotorVelocities().tolist())
        observation.extend(self.GetMotorTorques().tolist())
        observation.extend(list(self.GetBaseOrientation()))
        return observation

    def apply_action(self, motor_commands):
        """Set the desired motor angles to the motors of the minitaur.

        Note (hzyjerry): motor commands are set based on desired angles, not torques

        The desired motor angles are clipped based on the maximum allowed velocity.
        If the pd_control_enabled is True, a torque is calculated according to
        the difference between current and desired joint angle, as well as the joint
        velocity. This torque is exerted to the motor. For more information about
        PD control, please refer to: https://en.wikipedia.org/wiki/PID_controller.

        Args:
          motor_commands: The eight desired motor angles.
        """
        #print("motor commands 1", motor_commands)
        if self.motor_velocity_limit < np.inf:
            current_motor_angle = self.GetMotorAngles()
            motor_commands_max = (
                current_motor_angle + self.time_step * self.motor_velocity_limit)
            motor_commands_min = (
                current_motor_angle - self.time_step * self.motor_velocity_limit)
            #motor_commands = np.clip(motor_commands, motor_commands_min, motor_commands_max)
        #print("motor commands 2", motor_commands)
        if self.accurate_motor_model_enabled or self.pd_control_enabled:
            q = self.GetMotorAngles()
            qdot = self.GetMotorVelocities()
            if self.accurate_motor_model_enabled:
                actual_torque, observed_torque = self._motor_model.convert_to_torque(
                    motor_commands, q, qdot)
                #print("q", q, "qdot", qdot)
                #print("motor commands", motor_commands)
                #print("actual torque", actual_torque, "observed torque", observed_torque)
                self.debug_count += 1
                #if self.debug_count == 30:
                #    sys.exit()
                for i in range(len(self.qmax)):
                    if q[i] > self.qmax[i]:
                        self.qmax[i] = q[i]
                #print("Q max", self.qmax)
    
                if self.motor_overheat_protection:
                    for i in range(self.num_motors):
                        if abs(actual_torque[i]) > self.OVERHEAT_SHUTDOWN_TORQUE:
                            self._overheat_counter[i] += 1
                        else:
                            self._overheat_counter[i] = 0
                        if (self._overheat_counter[i] >
                            self.OVERHEAT_SHUTDOWN_TIME / self.time_step):
                            self._motor_enabled_list[i] = False
                # The torque is already in the observation space because we use
                # GetMotorAngles and GetMotorVelocities.
                self.observed_motor_torques = observed_torque
                #actual_torque.fill(0.0)

                # Transform into the motor space when applying the torque.
                self.applied_motor_torques = np.multiply(actual_torque,
                                                         self.motor_direction)
                for motor_id, motor_torque, motor_enabled in zip(
                    self._motor_id_list, self.applied_motor_torques,
                    self._motor_enabled_list):
                    if motor_enabled:
                        self._SetMotorTorqueById(motor_id, motor_torque)
                    else:
                        self._SetMotorTorqueById(motor_id, 0)
                #print("Apply motor", self.applied_motor_torques)
                for i in range(len(self.fmax)):
                    if motor_commands[i] > self.fmax[i]:
                        self.fmax[i] = motor_commands[i]
                #print("F max", self.fmax)
    
            else:
                torque_commands = -self._kp * (q - motor_commands) - self._kd * qdot

                # The torque is already in the observation space because we use
                # GetMotorAngles and GetMotorVelocities.
                self.observed_motor_torques = torque_commands

                # Transform into the motor space when applying the torque.
                self.applied_motor_torques = np.multiply(self.observed_motor_torques,
                                                          self.motor_direction)

                for motor_id, motor_torque in zip(self._motor_id_list,
                                                  self.applied_motor_torques):
                    self._SetMotorTorqueById(motor_id, motor_torque)
                print("Apply motor", self.applied_motor_torques)
        else:
            motor_commands_with_direction = np.multiply(motor_commands,
                                                      self.motor_direction)
            for motor_id, motor_command_with_direction in zip(
                self._motor_id_list, motor_commands_with_direction):
                print("command", motor_command_with_direction)
                self._SetDesiredMotorAngleById(motor_id, motor_command_with_direction)

    def GetMotorAngles(self):
        """Get the eight motor angles at the current moment.

        Returns:
          Motor angles.
        """
        motor_angles = [
            p.getJointState(self.minitaur, motor_id)[0]
            for motor_id in self._motor_id_list
        ]
        motor_angles = np.multiply(motor_angles, self.motor_direction)
        return motor_angles

    def GetMotorVelocities(self):
        """Get the velocity of all eight motors.

        Returns:
          Velocities of all eight motors.
        """
        motor_velocities = [
            p.getJointState(self.minitaur, motor_id)[1]
            for motor_id in self._motor_id_list
        ]
        motor_velocities = np.multiply(motor_velocities, self.motor_direction)
        return motor_velocities

    def GetMotorTorques(self):
        """Get the amount of torques the motors are exerting.

        Returns:
          Motor torques of all eight motors.
        """
        if self.accurate_motor_model_enabled or self.pd_control_enabled:
            return self.observed_motor_torques
        else:
            motor_torques = [
                p.getJointState(self.minitaur, motor_id)[3] for motor_id in self._motor_id_list ]
            motor_torques = np.multiply(motor_torques, self.motor_direction)
        return motor_torques

    def ConvertFromLegModel(self, actions):
        """Convert the actions that use leg model to the real motor actions.

        Args:
          actions: The theta, phi of the leg model.
        Returns:
          The eight desired motor angles that can be used in ApplyAction().
        """
        motor_angle = copy.deepcopy(actions)
        scale_for_singularity = 1
        offset_for_singularity = 1.5
        half_num_motors = int(self.num_motors / 2)
        quater_pi = math.pi / 4
        for i in range(self.num_motors):
            action_idx = i // 2
            forward_backward_component = (-scale_for_singularity * quater_pi * (
                actions[action_idx + half_num_motors] + offset_for_singularity))
            extension_component = (-1)**i * quater_pi * actions[action_idx]
            if i >= half_num_motors:
                extension_component = -extension_component
            motor_angle[i] = (
                math.pi + forward_backward_component + extension_component)
        return motor_angle

    def GetBaseMassFromURDF(self):
        """Get the mass of the base from the URDF file."""
        return self._base_mass_urdf

    def GetLegMassesFromURDF(self):
        """Get the mass of the legs from the URDF file."""
        return self._leg_masses_urdf

    def SetBaseMass(self, base_mass):
        p.changeDynamics(
            self.minitaur, self.BASE_LINK_ID, mass=base_mass)

    def SetLegMasses(self, leg_masses):
        """Set the mass of the legs.

        A leg includes leg_link and motor. All four leg_links have the same mass,
        which is leg_masses[0]. All four motors have the same mass, which is
        leg_mass[1].

        Args:
          leg_masses: The leg masses. leg_masses[0] is the mass of the leg link.
            leg_masses[1] is the mass of the motor.
        """
        for link_id in self.LEG_LINK_ID:
            p.changeDynamics(
                self.minitaur, link_id, mass=leg_masses[0])
        for link_id in self.MOTOR_LINK_ID:
            p.changeDynamics(
                self.minitaur, link_id, mass=leg_masses[1])

    def SetFootFriction(self, foot_friction):
        """Set the lateral friction of the feet.

        Args:
          foot_friction: The lateral friction coefficient of the foot. This value is
            shared by all four feet.
        """
        for link_id in self.FOOT_LINK_ID:
            p.changeDynamics(
                self.minitaur, link_id, lateralFriction=foot_friction)

    def SetBatteryVoltage(self, voltage):
        if self.accurate_motor_model_enabled:
            self._motor_model.set_voltage(voltage)

    def SetMotorViscousDamping(self, viscous_damping):
        if self.accurate_motor_model_enabled:
            self._motor_model.set_viscous_damping(viscous_damping)



class Minitaur(MinitaurBase):
    '''Wrapper class for gibson interface
    
    Attribtues:
        self.eyes
        self.resolution
        self.walk_target_x, self.walk_target_y
        self.mjcf_scaling
        self.observation_space
        self.action_space
        self.sensor_space

    Interface:
        self.apply_action()
        self.calc_state()
        self.addToScene()
    '''
    def __init__(self, config, env, pd_control_enabled=True, accurate_motor_model_enabled=True):
        MinitaurBase.__init__(self, config, env, pd_control_enabled, accurate_motor_model_enabled)

    def calc_state(self):
        MinitaurBase.GetObservation(self)


import sys, os
sys.path.append(os.path.dirname(__file__))
import pybullet as p
import time
import gym


class Scene:
    "A base class for single- and multiplayer scenes"

    def __init__(self, gravity, timestep, frame_skip, env):
        #self.np_random, seed = gym.utils.seeding.np_random(None)
        self.timestep = timestep
        self.frame_skip = frame_skip
        self.env = env

        self.dt = self.timestep * self.frame_skip
        self.cpp_world = World(gravity, timestep, frame_skip)

        self.test_window_still_open = True  # or never opened
        self.human_render_detected = False  # if user wants render("human"), we open test window

        self.multiplayer_robots = {}

    def test_window(self):
        "Call this function every frame, to see what's going on. Not necessary in learning."
        self.human_render_detected = True
        return self.test_window_still_open

    def actor_introduce(self, robot):
        "Usually after scene reset"
        if not self.multiplayer: return
        self.multiplayer_robots[robot.player_n] = robot

    def actor_is_active(self, robot):
        """
        Used by robots to see if they are free to exclusiveley put their HUD on the test window.
        Later can be used for click-focus robots.
        """
        return not self.multiplayer

    def episode_restart(self):
        "This function gets overridden by specific scene, to reset specific objects into their start positions"
        self.cpp_world.clean_everything()
        #self.cpp_world.test_window_history_reset()

    def global_step(self):
        """
        The idea is: apply motor torques for all robots, then call global_step(), then collect
        observations from robots using step() with the same action.
        """
        self.cpp_world.step(self.frame_skip)

class SingleRobotEmptyScene(Scene):
    multiplayer = False  # this class is used "as is" for InvertedPendulum, Reacher

class World:

    def __init__(self, gravity, timestep, frame_skip):
        self.gravity = gravity
        self.timestep = timestep
        self.frame_skip = frame_skip
        self.clean_everything()
        self._is_render = False
        self._last_frame_time = 0.0

    def clean_everything(self):
        #p.resetSimulation()
        p.setGravity(0, 0, -self.gravity)
        #p.setDefaultContactERP(0.9)
        p.setPhysicsEngineParameter(fixedTimeStep=self.timestep*self.frame_skip, numSolverIterations=50, numSubSteps=(self.frame_skip-1))

    def step(self, frame_skip):
        if self._is_render:
            # Sleep, otherwise the computation takes less time than real time,
            # which will make the visualization like a fast-forward video.
            time_spent = time.time() - self._last_frame_time
            self._last_frame_time = time.time()
            #time_to_sleep = self.timestep * self.frame_skip - time_spent
            #if time_to_sleep > 0:
            #    time.sleep(time_to_sleep)
        p.stepSimulation()





import numpy as np
import gibson.core.physics.settings as settings
from transforms3d import euler, quaternions


class PhysicsObject(object):
    """
    Controllable object in world
    This class mostly handles action parsing

    Note: only stores realtime delta pose, absolute pose is not
    stored

    Note: conventions
    By OpenGL convention, object is default: +x facing,
    which is incompatible with camera view matrix (-z facing).
    The inconsistency is handled by _cameraCalibrate function.
    By default, internal quaternion variables are [x, y, z, w]
    """

    def __init__(self, uid, simulator, pos, quat, v_t, v_r, fps):
        self.uid = uid
        self.sim = simulator

        self.v_t = float(v_t)
        self.v_r = float(v_r)
        self.fps = float(fps)
        self.action = self._createDefaultAction()

        #self.camera_offset  = np.array([0, 0, 0.8])
        self.camera_offset  = np.array([0, 0, 0])

        self.pos_init_xyz   = np.array(pos)
        self.quat_init_xyzw = np.array(quat)

        ## Relative delta rotation of object to world
        self.d_alpha, self.d_beta, self.d_gamma = 0, 0, 0

        ## Relative delta position inside object's world view
        self.d_xyz = np.array([0, 0, 0], dtype=float)
        self._updateInitialPositionOrientation()

        ## DEPRECATED: roll, pitch, yaw
        self.roll, self.pitch, self.yaw = 0, np.pi/6, 0


    def _updateInitialPositionOrientation(self):
        """ Update initial physics simulation
        Similar to implementation in updatePositionOrientation()
        """
        pos_world_xyz   = self.pos_init_xyz
        quat_world_xyzw = self.quat_init_xyzw

        ## parse xyz, alpha, beta, gamma from world
        ## apply local delta xyz, delta alpha, delta beta, delta gamma
        ## update to world
        quat_world_xyzw = self._rotateIntrinsic(quat_world_xyzw)
        ## Calibrate
        quat_world_xyzw = self._cameraCalibrate(quat_world_xyzw)
        #pos_xyz   = self._translateIntrinsic()
        self.sim.resetBasePositionAndOrientation(self.uid, pos_world_xyz, quat_world_xyzw)


    def updatePositionOrientation(self):
        """ Update physics simulation
        Physics simulation = object position + object rotation

        Note: When using sim.stepSimulation(), you need to call
        updatePositionOrientation() periodically.
        """
        pos_world_xyz, quat_world_xyzw = self.sim.getBasePositionAndOrientation(self.uid)
        # velocity
        v_translate, v_rotate = self.sim.getBaseVelocity(self.uid)

        quat_world_xyzw     = self._cameraUncalibrate(quat_world_xyzw)

        new_pos_world_xyz   = self._translateIntrinsic(pos_world_xyz, quat_world_xyzw)
        new_quat_world_xyzw = self._rotateIntrinsic(quat_world_xyzw)
        ## Calibrate
        new_quat_world_xyzw = self._cameraCalibrate(new_quat_world_xyzw)

        self.sim.resetBasePositionAndOrientation(self.uid, new_pos_world_xyz, new_quat_world_xyzw)
        self.sim.resetBaseVelocity(self.uid, v_translate, v_rotate)


    def getViewPosAndOrientation(self):
        """Output real-time pose for view renderer

        Note: the output pose & orientation are absolute pose defined
        inside physics world coordinate. The viewer might use different
        convention, e.g. a coordinate relative to initial pose.
        PhysicsObject is agnostic to viewer convention. It is the viewer's
        job to handle this input/output
        """
        pos_world_xyz, quat_world_xyzw = self.sim.getBasePositionAndOrientation(self.uid)

        pos_world_xyz         = np.array(pos_world_xyz)
        quat_world_xyzw     = np.array(quat_world_xyzw)

        quat_world_xyzw     = self._cameraUncalibrate(quat_world_xyzw)
        quat_world_wxyz     = PhysicsObject.quatXyzwToWxyz(quat_world_xyzw)

        """
        quat_init_wxyz     = PhysicsObject.quatXyzwToWxyz(self.quat_init_xyzw)
        pos_view_xyz     = pos_world_xyz - self.pos_init_xyz
        quat_view_wxyz     = quaternions.qmult(quaternions.qinverse(quat_init_wxyz), quat_world_wxyz)
        euler_view       = euler.quat2euler(quat_view_wxyz)
        """

        pos_world_xyz = self._applyCameraOffset(pos_world_xyz)

        return pos_world_xyz.tolist(), quat_world_wxyz.tolist()



    def getUpdateFromKeyboard(self, restart=False):
        # Special Controls: B3G_RIGHT_ARROW, B3G_LEFT_ARROW,
        #     B3G_DOWN_ARROW, B3G_UP_ARROW

        self.action = self._createDefaultAction()
        keys = self.sim.getKeyboardEvents()


        if (ord('r') in keys or restart):
            self.action['restart'] = True
        if (ord('d') in keys):
            self.action['right'] = True
        if (ord('a') in keys):
            self.action['left'] = True
        if (ord('s') in keys):
            self.action['backward'] = True
        if (ord('w') in keys or ord('q') in keys):
            self.action['forward'] = True
        if (ord('z') in keys):
            self.action['up'] = True
        if (ord('c') in keys):
            self.action['down'] = True

        if (ord('u') in keys):
            self.action['alpha'] = 1
        if (ord('j') in keys):
            self.action['alpha'] = -1
        if (ord('i') in keys):
            self.action['beta'] = 1
        if (ord('k') in keys):
            self.action['beta'] = -1
        if (ord('o') in keys):
            self.action['gamma'] = 1
        if (ord('l') in keys):
            self.action['gamma'] = -1
        return keys
        #self.parseActionAndUpdate()


    def parseActionAndUpdate(self, action=None):
        """ Update position: because the object's rotation
        changes every time, the position needs to be updated
        by delta
        """
        if action:
            self.action = self._createDefaultAction()
            for k in action.keys():
                self.action[k] = action[k]
        delta_xyz = np.array([0, 0, 0], dtype=float)
        if self.action['restart']:
            self._restartLocationOrientation()
            return
        if self.action['up']:
            delta_xyz[1] =  self.v_t/settings.STEPS_PER_SEC
        if self.action['down']:
            delta_xyz[1] = -self.v_t/settings.STEPS_PER_SEC
        if self.action['left']:
            delta_xyz[0] = -self.v_t/settings.STEPS_PER_SEC
        if self.action['right']:
            delta_xyz[0] =  self.v_t/settings.STEPS_PER_SEC
        if self.action['forward']:
            delta_xyz[2] = -self.v_t/settings.STEPS_PER_SEC
        if self.action['backward']:
            delta_xyz[2] =  self.v_t/settings.STEPS_PER_SEC
        self.d_xyz = delta_xyz

        ## Update rotation: reset the rotation every time
        if self.action['alpha'] > 0:
            self.d_alpha =  self.v_r/settings.STEPS_PER_SEC
        if self.action['alpha'] < 0:
            self.d_alpha = -self.v_r/settings.STEPS_PER_SEC
        if self.action['beta'] > 0:
            self.d_beta =   self.v_r/settings.STEPS_PER_SEC
        if self.action['beta'] < 0:
            self.d_beta =  -self.v_r/settings.STEPS_PER_SEC
        if self.action['gamma'] > 0:
            self.d_gamma =  self.v_r/settings.STEPS_PER_SEC
        if self.action['gamma'] < 0:
            self.d_gamma = -self.v_r/settings.STEPS_PER_SEC

        self.updatePositionOrientation()
        #self.clearUpDelta()

    @staticmethod
    def quatWxyzToXyzw(wxyz):
        """
        wxyz: transforms3s array format
        xyzw: pybullet format
        """
        return np.concatenate((wxyz[1:], wxyz[:1]))

    @staticmethod
    def quatXyzwToWxyz(xyzw):
        """
        wxyz: transforms3s array format
        xyzw: pybullet format
        """
        return np.concatenate((xyzw[-1:], xyzw[:-1]))


    def rotate_quat_by_euler(xyzw, e_x, e_y, e_z):
        """
        wxyz: transforms3s array format
        xyzw: pybullet format
        """
        wxyz = PhysicsObject.quatXyzwToWxyz(xyzw)
        rot_mat = euler.euler2mat(e_x, e_y, e_z)
        wxyz = quaternions.qmult(rot_mat, wxyz)
        return PhysicsObject.quatWxyzToXyzw(wxyz)


    def _applyCameraOffset(self, pos_xyz):
        pos_camera_xyz = pos_xyz + self.camera_offset
        return pos_camera_xyz


    def _cameraCalibrate(self, org_quat_xyzw):
        """ Convert object's head from +x facing to -z facing
        Note that this is not needed when you're computing view_matrix,
        only use this function for adjusting object head
        To get object rotation at current object pose:
            rotation = self._cameraCalibrate(self._rotateIntrinsic(
                   self.quat_world))
        To get view rotation at current object pose:
              rotation = self._rotateIntrinsic(self.quat_world)
          """
        z_facing_wxyz = euler.euler2quat(-np.pi/2, np.pi/2, 0)
        org_quat_wxyz = PhysicsObject.quatXyzwToWxyz(org_quat_xyzw)
        new_quat_xyzw = PhysicsObject.quatWxyzToXyzw(quaternions.qmult(org_quat_wxyz, z_facing_wxyz))
        return new_quat_xyzw


    def _createDefaultAction(self):
        action = {
            'up'      : False,
            'down'      : False,
            'left'      : False,
            'right'      : False,
            'forward' : False,
            'backward': False,
            'restart' : False,
            'alpha'   : 0,
            'beta'    : 0,
            'gamma'   : 0
        }
        return action

    def _cameraUncalibrate(self, new_quat_xyzw):
        """ Undo the effect of _cameraCalibrate
        Needed in two cases:
        (1) Updating pose in physics engine

            self._cameraUnalibrate(sim.getBasePositionAndOrientation(uid))

        This is because object orientation in physics engine has
        been calibrated. To update parameter (alpha, beta, gamma),
        need to uncalibrate first. See updatePositionOrientation()
          (2) Sending pose to viewer renderer

             send(self._cameraUnalibrate(sim.See updatePositionOrientation(uid)))

        This is because we need to send uncalibrated view pose
        """
        x_facing_wxyz = euler.euler2quat(np.pi/2, 0, -np.pi/2)
        new_quat_wxyz = PhysicsObject.quatXyzwToWxyz(new_quat_xyzw)
        org_quat_wxyz = PhysicsObject.quatWxyzToXyzw(quaternions.qmult(new_quat_wxyz, x_facing_wxyz))
        return org_quat_wxyz


    def _translateIntrinsic(self, pos_world_xyz, quat_world_xyzw):
        """ Add intrinsic translation to extrinsic
        Extrinsic translation = Intrinsic delta translation +
            Extrinsic translation

        Note: order doesn't matter
        """
        delta_objec_xyz = self.d_xyz
        quat_world_wxyz = PhysicsObject.quatXyzwToWxyz(quat_world_xyzw)
        delta_world_xyz = quaternions.quat2mat(quat_world_wxyz).dot(delta_objec_xyz)
        return pos_world_xyz + delta_world_xyz


    def _rotateIntrinsic(self, quat_world_xyzw):
        """ Add intrinsic rotation to extrinsic
        Extrinsic rotation = Extrinsic rotation *
            Intrinsic delta rotation

        Note: order matters. Apply intrinsic delta quat first, then
        extrinsic rotation
        """
        quat_world_wxyz = PhysicsObject.quatXyzwToWxyz(quat_world_xyzw)
        euler_world     = euler.quat2euler(quat_world_wxyz)

        quat_objec_wxyz = euler.euler2quat(self.d_alpha, self.d_beta, self.d_gamma)
        new_quat_world_wxyz  = quaternions.qmult(quat_world_wxyz, quat_objec_wxyz)
        new_quat_world_xyzw  = PhysicsObject.quatWxyzToXyzw(new_quat_world_wxyz)
        return new_quat_world_xyzw


    def _restartLocationOrientation(self):
        self._updateInitialPositionOrientation()


    def clearUpDelta(self):
        self.d_xyz = np.array([0, 0, 0], dtype=float)
        self.d_alpha, self.d_beta, self.d_gamma = 0, 0, 0


    ## DEPRECATED: roll, pitch, yaw
    def _principle_to_mat(self):
        alpha = self.yaw
        beta  = self.pitch
        gamma = self.roll
        mat   = np.eye(4)
        mat[0, 0] = cos(alpha)*cos(beta)
        mat[1, 0] = sin(alpha)*cos(beta)
        mat[2, 0] = -sin(beta)

        mat[0, 1] = cos(alpha)*sin(beta)*sin(gamma) - sin(alpha)*cos(gamma)
        mat[1, 1] = sin(alpha)*sin(beta)*sin(gamma) + cos(alpha)*cos(gamma)
        mat[2, 1] = cos(beta)*sin(gamma)

        mat[0, 2] = cos(alpha)*sin(beta)*cos(gamma) + sin(alpha)*sin(gamma)
        mat[1, 2] = sin(alpha)*sin(beta)*cos(gamma) - cos(alpha)*sin(gamma)
        mat[2, 2] = cos(beta)*cos(gamma)
        return mat



from gibson.envs.env_modalities import CameraRobotEnv, BaseRobotEnv, SemanticRobotEnv
from gibson.envs.env_bases import *
from gibson.core.physics.robot_locomotors import Husky
from transforms3d import quaternions
import os
import numpy as np
import sys
import pybullet as p
from gibson.core.physics.scene_stadium import SinglePlayerStadiumScene
import pybullet_data
import cv2

CALC_OBSTACLE_PENALTY = 0

tracking_camera = {
    'yaw': 110,
    'z_offset': 0.5,
    'distance': 1,
    'pitch': -20
}

tracking_camera_top = {
    'yaw': 20,  # demo: living room, stairs
    'z_offset': 0.5,
    'distance': 1,
    'pitch': -20
}

class HuskyNavigateEnv(CameraRobotEnv):
    """Specfy navigation reward
    """
    def __init__(self, config, gpu_idx=0):
        self.config = self.parse_config(config)
        assert(self.config["envname"] == self.__class__.__name__ or self.config["envname"] == "TestEnv")

        CameraRobotEnv.__init__(self, self.config, gpu_idx,
                                scene_type="stadium" if self.config["model_id"]=="stadium" else "building",
                                tracking_camera=tracking_camera)

        self.robot_introduce(Husky(self.config, env=self))
        self.scene_introduce()
        self.total_reward = 0
        self.total_frame = 0

    def add_text(self, img):
        font = cv2.FONT_HERSHEY_SIMPLEX
        x,y,z = self.robot.get_position()
        r,p,ya = self.robot.get_rpy()
        cv2.putText(img, 'x:{0:.4f} y:{1:.4f} z:{2:.4f}'.format(x,y,z), (10, 20), font, 0.5, (255, 255, 255), 1, cv2.LINE_AA)
        cv2.putText(img, 'ro:{0:.4f} pth:{1:.4f} ya:{2:.4f}'.format(r,p,ya), (10, 40), font, 0.5, (255, 255, 255), 1, cv2.LINE_AA)
        cv2.putText(img, 'potential:{0:.4f}'.format(self.potential), (10, 60), font, 0.5, (255, 255, 255), 1, cv2.LINE_AA)
        cv2.putText(img, 'fps:{0:.4f}'.format(self.fps), (10, 80), font, 0.5, (255, 255, 255), 1, cv2.LINE_AA)
        return img

    def _rewards(self, action=None, debugmode=False):
        a = action
        potential_old = self.potential
        self.potential = self.robot.calc_potential()
        progress = float(self.potential - potential_old)

        feet_collision_cost = 0.0
        for i, f in enumerate(
                self.robot.feet):  # TODO: Maybe calculating feet contacts could be done within the robot code
            contact_ids = set((x[2], x[4]) for x in f.contact_list())
            #contact_ids = set([x[2] for x in f.contact_list()])
            if (self.ground_ids & contact_ids):
                # see Issue 63: https://github.com/openai/roboschool/issues/63
                # feet_collision_cost += self.foot_collision_cost
                self.robot.feet_contact[i] = 1.0
            else:
                self.robot.feet_contact[i] = 0.0
        
        electricity_cost  = self.electricity_cost  * float(np.abs(a*self.robot.joint_speeds).mean())
        electricity_cost  += self.stall_torque_cost * float(np.square(a).mean())


        steering_cost = self.robot.steering_cost(a)
        debugmode = 0
        if debugmode:
            print("steering cost", steering_cost)

        wall_contact = []
        
        for i, f in enumerate(self.parts):
            if self.parts[f] not in self.robot.feet:
                wall_contact += [pt for pt in self.robot.parts[f].contact_list() if pt[6][2] > 0.15]
        debugmode = 0
        if debugmode:
            print("Husky wall contact:", len(wall_contact))
        wall_collision_cost = self.wall_collision_cost * len(wall_contact)

        joints_at_limit_cost = float(self.joints_at_limit_cost * self.robot.joints_at_limit)
        close_to_target = 0

        if self.robot.dist_to_target() < 2:
            close_to_target = 0.5

        angle_cost = self.robot.angle_cost()

        obstacle_penalty = 0
        if CALC_OBSTACLE_PENALTY and self._require_camera_input:
            obstacle_penalty = get_obstacle_penalty(self.robot, self.render_depth)

        debugmode = 0
        if debugmode:
            print("angle cost", angle_cost)

        height = self.robot.get_position()[2]
        pitch = self.robot.get_rpy()[1]
        alive = float(self.robot.alive_bonus(height, pitch))
        
        debugmode = 0
        if (debugmode):
            #print("Wall contact points", len(wall_contact))
            print("Collision cost", wall_collision_cost)
            #print("electricity_cost", electricity_cost)
            print("close to target", close_to_target)
            print("Obstacle penalty", obstacle_penalty)
            print("Steering cost", steering_cost)
            print("progress", progress)
            #print("electricity_cost")
            #print(electricity_cost)
            #print("joints_at_limit_cost")
            #print(joints_at_limit_cost)
            #print("feet_collision_cost")
            #print(feet_collision_cost)

        rewards = [
            #alive,
            progress,
            wall_collision_cost,
            close_to_target,
            steering_cost,
            #angle_cost,
            #obstacle_penalty
            #electricity_cost,
            #joints_at_limit_cost,
            #feet_collision_cost
        ]
        return rewards

    def _termination(self, debugmode=False):
        height = self.robot.get_position()[2]
        pitch = self.robot.get_rpy()[1]
        alive = float(self.robot.alive_bonus(height, pitch)) > 0
        #alive = len(self.robot.parts['top_bumper_link'].contact_list()) == 0

        done = not alive or self.nframe > 250 or height < 0
        #if done:
        #    print("Episode reset")
        return done

    def _flag_reposition(self):
        target_pos = self.robot.target_pos

        self.flag = None
        if self.gui and not self.config["display_ui"]:
            self.visual_flagId = p.createVisualShape(p.GEOM_MESH, fileName=os.path.join(pybullet_data.getDataPath(), 'cube.obj'), meshScale=[0.5, 0.5, 0.5], rgbaColor=[1, 0, 0, 0.7])
            self.last_flagId = p.createMultiBody(baseVisualShapeIndex=self.visual_flagId, baseCollisionShapeIndex=-1, basePosition=[target_pos[0], target_pos[1], 0.5])

    def  _reset(self):
        self.total_frame = 0
        self.total_reward = 0
        obs = CameraRobotEnv._reset(self)
        self._flag_reposition()
        return obs

    ## openai-gym v0.10.5 compatibility
    step  = CameraRobotEnv._step



class HuskyNavigateSpeedControlEnv(HuskyNavigateEnv):
    """Specfy navigation reward
    """
    def __init__(self, config, gpu_idx=0):
        #assert(self.config["envname"] == self.__class__.__name__ or self.config["envname"] == "TestEnv")
        HuskyNavigateEnv.__init__(self, config, gpu_idx)
        self.robot.keys_to_action = {
            (ord('s'), ): [-0.5,0], ## backward
            (ord('w'), ): [0.5,0], ## forward
            (ord('d'), ): [0,-0.5], ## turn right
            (ord('a'), ): [0,0.5], ## turn left
            (): [0,0]
        }

        self.base_action_omage = np.array([-0.001, 0.001, -0.001, 0.001])
        self.base_action_v = np.array([0.001, 0.001, 0.001, 0.001])
        self.action_space = gym.spaces.Discrete(5)
        #control_signal = -0.5
        #control_signal_omega = 0.5
        self.v = 0
        self.omega = 0
        self.kp = 100
        self.ki = 0.1
        self.kd = 25
        self.ie = 0
        self.de = 0
        self.olde = 0
        self.ie_omega = 0
        self.de_omega = 0
        self.olde_omage = 0


    def _step(self, action):
        control_signal, control_signal_omega = action
        self.e = control_signal - self.v
        self.de = self.e - self.olde
        self.ie += self.e
        self.olde = self.e
        pid_v = self.kp * self.e + self.ki * self.ie + self.kd * self.de

        self.e_omega = control_signal_omega - self.omega
        self.de_omega = self.e_omega - self.olde_omage
        self.ie_omega += self.e_omega
        pid_omega = self.kp * self.e_omega + self.ki * self.ie_omega + self.kd * self.de_omega

        obs, rew, env_done, info = HuskyNavigateEnv.step(self, pid_v * self.base_action_v + pid_omega * self.base_action_omage)

        self.v = obs["nonviz_sensor"][3]
        self.omega = obs["nonviz_sensor"][-1]

        return obs,rew,env_done,info

    ## openai-gym v0.10.5 compatibility
    step  = _step



class HuskyGibsonFlagRunEnv(CameraRobotEnv):
    """Specfy flagrun reward
    """
    def __init__(self, config, gpu_idx=0):
        self.config = self.parse_config(config)
        print(self.config["envname"])
        assert(self.config["envname"] == self.__class__.__name__ or self.config["envname"] == "TestEnv")
        CameraRobotEnv.__init__(self, self.config, gpu_idx,
                                scene_type="building",
                                tracking_camera=tracking_camera)

        self.robot_introduce(Husky(self.config, env=self))
        self.scene_introduce()

        self.total_reward = 0
        self.total_frame = 0
        self.flag_timeout = 1
        self.visualid = -1
        self.lastid = None
        self.gui = self.config["mode"] == "gui"
        
        if self.gui:
            self.visualid = p.createVisualShape(p.GEOM_MESH, fileName=os.path.join(pybullet_data.getDataPath(), 'cube.obj'), meshScale=[0.2, 0.2, 0.2], rgbaColor=[1, 0, 0, 0.7])
        self.colisionid = p.createCollisionShape(p.GEOM_MESH, fileName=os.path.join(pybullet_data.getDataPath(), 'cube.obj'), meshScale=[0.2, 0.2, 0.2])

        self.lastid = None
        self.obstacle_dist = 100
        
    def _reset(self):
        obs = CameraRobotEnv._reset(self)
        return obs

    def _flag_reposition(self):
        #self.walk_target_x = self.np_random.uniform(low=-self.scene.stadium_halflen,
        #                                            high=+self.scene.stadium_halflen)
        #self.walk_target_y = self.np_random.uniform(low=-self.scene.stadium_halfwidth,
        #                                            high=+self.scene.stadium_halfwidth)
        force_x = self.np_random.uniform(-300,300)
        force_y = self.np_random.uniform(-300, 300)

        more_compact = 0.5  # set to 1.0 whole football field
        #self.walk_target_x *= more_compact
        #self.walk_target_y *= more_compact

        startx, starty, _ = self.robot.get_position()


        self.flag = None
        #self.flag = self.scene.cpp_world.debug_sphere(self.walk_target_x, self.walk_target_y, 0.2, 0.2, 0xFF8080)
        self.flag_timeout = 3000 / self.scene.frame_skip
        #print('targetxy', self.flagid, self.walk_target_x, self.walk_target_y, p.getBasePositionAndOrientation(self.flagid))
        #p.resetBasePositionAndOrientation(self.flagid, posObj = [self.walk_target_x, self.walk_target_y, 0.5], ornObj = [0,0,0,0])
        if self.lastid:
            p.removeBody(self.lastid)

        self.lastid = p.createMultiBody(baseMass = 1, baseVisualShapeIndex=self.visualid, baseCollisionShapeIndex=self.colisionid, basePosition=[startx, starty, 0.5])
        p.applyExternalForce(self.lastid, -1, [force_x,force_y,50], [0,0,0], p.LINK_FRAME)

        ball_xyz, _ = p.getBasePositionAndOrientation(self.lastid)

        self.robot.walk_target_x = ball_xyz[0]
        self.robot.walk_target_y = ball_xyz[1]

    def _rewards(self, action = None, debugmode=False):
        a = action
        potential_old = self.potential
        self.potential = self.robot.calc_potential()
        if self.flag_timeout > 225:
            progress = 0
        else:
            progress = float(self.potential - potential_old)

        if not a is None:
            electricity_cost = self.electricity_cost * float(np.abs(
                a * self.robot.joint_speeds).mean())  # let's assume we have DC motor with controller, and reverse current braking
            electricity_cost += self.stall_torque_cost * float(np.square(a).mean())
        else:
            electricity_cost = 0

        alive = len(self.robot.parts['top_bumper_link'].contact_list())
        if alive == 0:
            alive_score = 0.1
        else:
            alive_score = -0.1

        joints_at_limit_cost = float(self.joints_at_limit_cost * self.robot.joints_at_limit)
        debugmode = 0
        if (debugmode):
            print("progress")
            print(progress)

        obstacle_penalty = 0

        #print("obs dist %.3f" %self.obstacle_dist)
        if self.obstacle_dist < 0.7:
            obstacle_penalty = self.obstacle_dist - 0.7

        rewards = [
            alive_score,
            progress,
            obstacle_penalty
        ]
        return rewards

    def _termination(self, debugmode=False):
        alive = len(self.robot.parts['top_bumper_link'].contact_list())
        done = alive > 0 or self.nframe > 500
        if (debugmode):
            print("alive=")
            print(alive)
        print(len(self.robot.parts['top_bumper_link'].contact_list()), self.nframe, done)
        return done

    def _step(self, a):
        state, reward, done, meta = CameraRobotEnv._step(self, a)
        if self.flag_timeout <= 0 or (self.flag_timeout < 225 and self.robot.walk_target_dist < 0.8):
            self._flag_reposition()
        self.flag_timeout -= 1

        if "depth" in self.config["output"]:
            depth_obs = self.get_observations()["depth"]
            x_start = int(self.windowsz/2-16)
            x_end   = int(self.windowsz/2+16)
            y_start = int(self.windowsz/2-16)
            y_end   = int(self.windowsz/2+16)
            self.obstacle_dist = (np.mean(depth_obs[x_start:x_end, y_start:y_end, -1]))

        return state, reward, done, meta

    ## openai-gym v0.10.5 compatibility
    step  = _step


class HuskySemanticNavigateEnv(SemanticRobotEnv):
    """Specfy navigation reward
    """
    def __init__(self, config, gpu_idx=0):
        #assert(self.config["envname"] == self.__class__.__name__ or self.config["envname"] == "TestEnv")
        self.config = self.parse_config(config)
        SemanticRobotEnv.__init__(self, self.config, gpu_idx,
                                  scene_type="building",
                                  tracking_camera=tracking_camera)
        self.robot_introduce(Husky(self.config, env=self))
        self.scene_introduce()

        self.total_reward = 0
        self.total_frame = 0
        self.flag_timeout = 1
        self.visualid = -1
        self.lastid = None
        self.gui = self.config["mode"] == "gui"

        if self.gui:
            self.visualid = p.createVisualShape(p.GEOM_MESH, fileName=os.path.join(pybullet_data.getDataPath(), 'cube.obj'), meshScale=[0.2, 0.2, 0.2], rgbaColor=[1, 0, 0, 0.7])
        self.colisionid = p.createCollisionShape(p.GEOM_MESH, fileName=os.path.join(pybullet_data.getDataPath(), 'cube.obj'), meshScale=[0.2, 0.2, 0.2])

        self.lastid = None
        self.obstacle_dist = 100

        self.semantic_flagIds = []

        debug_semantic = 1
        if debug_semantic and self.gui:
            for i in range(self.semantic_pos.shape[0]):
                pos = self.semantic_pos[i]
                pos[2] += 0.2   # make flag slight above object 
                visualId = p.createVisualShape(p.GEOM_MESH, fileName=os.path.join(pybullet_data.getDataPath(), 'cube.obj'), meshScale=[0.1, 0.1, 0.1], rgbaColor=[1, 0, 0, 0.7])
                flagId = p.createMultiBody(baseVisualShapeIndex=visualId, baseCollisionShapeIndex=-1, basePosition=pos)
                self.semantic_flagIds.append(flagId)

    def step(self, action):
        obs, rew, env_done, info = SemanticRobotEnv.step(self,action=action)
        self.close_semantic_ids = self.get_close_semantic_pos(dist_max=1.0, orn_max=np.pi/5)
        for i in self.close_semantic_ids:
            flagId = self.semantic_flagIds[i]
            p.changeVisualShape(flagId, -1, rgbaColor=[0, 1, 0, 1])
        return obs,rew,env_done,info

    def _rewards(self, action = None, debugmode=False):
        a = action
        potential_old = self.potential
        self.potential = self.robot.calc_potential()
        if self.flag_timeout > 225:
            progress = 0
        else:
            progress = float(self.potential - potential_old)

        if not a is None:
            electricity_cost = self.electricity_cost * float(np.abs(
                a * self.robot.joint_speeds).mean())  # let's assume we have DC motor with controller, and reverse current braking
            electricity_cost += self.stall_torque_cost * float(np.square(a).mean())
        else:
            electricity_cost = 0

        alive = len(self.robot.parts['top_bumper_link'].contact_list())
        if alive == 0:
            alive_score = 0.1
        else:
            alive_score = -0.1

        joints_at_limit_cost = float(self.joints_at_limit_cost * self.robot.joints_at_limit)
        debugmode = 0
        if (debugmode):
            print("progress")
            print(progress)

        obstacle_penalty = 0

        #print("obs dist %.3f" %self.obstacle_dist)
        if self.obstacle_dist < 0.7:
            obstacle_penalty = self.obstacle_dist - 0.7

        rewards = [
            alive_score,
            progress,
            obstacle_penalty
        ]
        return rewards

    def _termination(self, debugmode=False):
        alive = len(self.robot.parts['top_bumper_link'].contact_list())
        done = alive > 0 or self.nframe > 500
        if (debugmode):
            print("alive=")
            print(alive)
        #print(len(self.robot.parts['top_bumper_link'].contact_list()), self.nframe, done)
        return done

    def _reset(self):
        CameraRobotEnv._reset(self)
        for flagId in self.semantic_flagIds:
            p.changeVisualShape(flagId, -1, rgbaColor=[1, 0, 0, 1])


def get_obstacle_penalty(robot, depth):
    screen_sz = robot.obs_dim[0]
    screen_delta = int(screen_sz / 8)
    screen_half  = int(screen_sz / 2)
    height_offset = int(screen_sz / 4)

    obstacle_dist = (np.mean(depth[screen_half  + height_offset - screen_delta : screen_half + height_offset + screen_delta, screen_half - screen_delta : screen_half + screen_delta, -1]))
    obstacle_penalty = 0
    OBSTACLE_LIMIT = 1.5
    if obstacle_dist < OBSTACLE_LIMIT:
       obstacle_penalty = (obstacle_dist - OBSTACLE_LIMIT)
    
    debugmode = 0
    if debugmode:
        #print("Obstacle screen", screen_sz, screen_delta)
        print("Obstacle distance", obstacle_dist)
        print("Obstacle penalty", obstacle_penalty)
    return obstacle_penalty


from gibson.data.datasets import ViewDataSet3D, get_model_path
from gibson.core.render.pcrender import PCRenderer
from gibson.core.render.profiler import Profiler
from gibson.envs.env_bases import BaseEnv
from gibson.envs.env_utils import *
import gibson
from gym import error
from gym.utils import seeding
from transforms3d import quaternions
from gibson.envs.env_ui import *
import pybullet as p
import pybullet_data
from tqdm import *
import subprocess, os, signal
import numpy as np
import sys
import zmq
import pygame
from pygame import surfarray
import socket
import shlex
import gym
import cv2
import os.path as osp
import os
from PIL import Image

from transforms3d.euler import euler2quat, euler2mat
from transforms3d.quaternions import quat2mat, qmult
import transforms3d.quaternions as quat
import time

DEFAULT_TIMESTEP  = 1.0/(4 * 9)
DEFAULT_FRAMESKIP = 4
DEFAULT_DEBUG_CAMERA = {
    'yaw': 30,
    'distance': 2.5,
    'pitch': -35,
    'z_offset': 0
}

DEPTH_SCALE_FACTOR = 35
DEPTH_OFFSET_FACTOR = 20

class BaseRobotEnv(BaseEnv):
    """Based on BaseEnv
    Handles action, reward
    """
    DEFAULT_PORT = 5556

    def __init__(self, config, tracking_camera, scene_type="building", gpu_idx=0):
        BaseEnv.__init__(self, config, scene_type, tracking_camera)

        self.camera_x = 0
        self.walk_target_x = 1e3  # kilometer away
        self.walk_target_y = 0
        self.k = 5
        self.robot_tracking_id = -1

        self.scale_up  = 4
        self.dataset = None
        self.ground_ids = None
        if self.gui:
            assert(self.tracking_camera is not None)
        self.gpu_idx = gpu_idx
        self.assign_ports()
        self.nframe = 0
        self.eps_reward = 0

        self.reward = 0
        self.eps_count = 0

        self._robot_introduced = False
        self._scene_introduced = False

    def assign_ports(self):
        '''Rendering multiple modalities (RGB, depth, normal) needs to be done 
        on different ports. Assign individual ports to each modality:

        | Rendering | Port         |
        | RGB       | Default      |
        | Depth     | Default - 1  |
        | Normal    | Default - 2  |
        | Semantics | Default - 3  |
        | UI        | Default - 4  |
        Default depends on how many Gibson environments are running simultanously
        '''
        self.port_rgb = self.DEFAULT_PORT - self.gpu_idx * 5
        self.port_depth = self.port_rgb - 1
        self.port_normal = self.port_rgb - 2
        self.port_sem  = self.port_rgb - 3
        self.port_ui   = self.port_rgb - 4

    def robot_introduce(self, robot):
        self.robot = robot
        self.robot.env = self
        self.action_space = self.robot.action_space
        ## Robot's eye observation, in sensor mode black pixels are returned
        self.observation_space = self.robot.observation_space
        self.sensor_space = self.robot.sensor_space
        # seed for robot
        self.robot.np_random = self.np_random
        self._robot_introduced = True
        assert (self.robot.resolution <= 512 and self.robot.resolution >= 64), \
            "Robot resolution must in [64, 512]"
        #if self.robot.resolution == 64:
        #    self.windowsz = 64
        #    self.scale_up = 4
        #elif self.robot.resolution == 128:
        #    self.windowsz = 128
        #    self.scale_up = 4
        #elif self.robot.resolution == 256:
        #    self.windowsz = 256
        #    self.scale_up = 2
        #else:
        #    self.windowsz = 512
        #    self.scale_up = 1

        self.windowsz = self.robot.resolution
        self.scale_up = int(512 / self.windowsz)

        if "fast_lq_render" in self.config and self.config["fast_lq_render"] == True:
            self.scale_up *= 2
        # if fast render, use lower quality point cloud

        self._render_width = self.windowsz
        self._render_height = self.windowsz

    def scene_introduce(self, gravity=9.8, collision_enabled=True):
        assert(self._robot_introduced)
        self.create_scene(gravity=gravity, collision_enabled=collision_enabled)
        self._scene_introduced = True

    def get_keys_to_action(self):
        return self.robot.keys_to_action

    def _reset(self):
        assert(self._robot_introduced)
        assert(self._scene_introduced)
        debugmode = 1
        if debugmode:
            print("Episode: steps:{} score:{}".format(self.nframe, self.reward))
            body_xyz = self.robot.body_xyz
            #print("[{}, {}, {}],".format(body_xyz[0], body_xyz[1], body_xyz[2]))
            print("Episode count: {}".format(self.eps_count))
            self.eps_count += 1
        self.nframe = 0
        self.eps_reward = 0
        BaseEnv._reset(self)

        if not self.ground_ids:
            self.parts, self.jdict, self.ordered_joints, self.robot_body = self.robot.addToScene(
                    [])
            self.ground_ids = set(self.scene.scene_obj_list)

        ## Todo: (hzyjerry) this part is not working, robot_tracking_id = -1
        for i in range (p.getNumBodies()):
            if (p.getBodyInfo(i)[0].decode() == self.robot_body.get_name()):
               self.robot_tracking_id=i
            #print(p.getBodyInfo(i)[0].decode())
        i = 0

        eye_pos, eye_quat = self.get_eye_pos_orientation()
        pose = [eye_pos, eye_quat]

        observations = self.render_observations(pose)
        pos = self.robot._get_scaled_position()
        orn = self.robot.get_orientation()

        pos = (pos[0], pos[1], pos[2] + self.tracking_camera['z_offset'])
        p.resetDebugVisualizerCamera(self.tracking_camera['distance'],self.tracking_camera['yaw'], self.tracking_camera['pitch'],pos)
        return observations

    electricity_cost     = -2.0 # cost for using motors -- this parameter should be carefully tuned against reward for making progress, other values less improtant
    stall_torque_cost   = -0.1  # cost for running electric current through a motor even at zero rotational speed, small
    foot_collision_cost  = -1.0 # touches another leg, or other objects, that cost makes robot avoid smashing feet into itself
    wall_collision_cost = -0.1
    foot_ground_object_names = set(["buildingFloor"])  # to distinguish ground and other objects
    joints_at_limit_cost = -0.1 # discourage stuck joints


    def _step(self, a):
        self.nframe += 1
        if not self.scene.multiplayer:  # if multiplayer, action first applied to all robots, then global step() called, then _step() for all robots with the same actions
            self.robot.apply_action(a)
            self.scene.global_step()

        self.rewards = self._rewards(a)
        done = self._termination()
        debugmode=0
        if (debugmode):
            print("rewards=")
            print(self.rewards)
            print("sum rewards")
            print(sum(self.rewards))

        self.reward += sum(self.rewards)
        self.eps_reward += sum(self.rewards)

        debugmode = 0
        if debugmode:
            print("Eps frame {} reward {}".format(self.nframe, self.reward))
            print("position", self.robot.get_position())
        if self.gui:
            pos = self.robot._get_scaled_position()
            orn = self.robot.get_orientation()
            pos = (pos[0], pos[1], pos[2] + self.tracking_camera['z_offset'])
            pos = np.array(pos)
            dist = self.tracking_camera['distance'] / self.robot.mjcf_scaling
            p.resetDebugVisualizerCamera(dist, self.tracking_camera['yaw'], self.tracking_camera['pitch'], pos)

        eye_pos, eye_quat = self.get_eye_pos_orientation()
        pose = [eye_pos, eye_quat]
        observations = self.render_observations(pose)

        debugmode = 0
        if (debugmode):
            print("Camera env eye position", eye_pos)
            print("episode rewards", sum(self.rewards), "steps", self.nframe)

        episode = None
        if done:
            episode = {'r': self.reward,
                       'l': self.nframe}
            debugmode = 0
            if debugmode:
                print("return episode:", episode)
        return observations, sum(self.rewards), bool(done), dict(eye_pos=eye_pos, eye_quat=eye_quat, episode=episode)

    def _termination(self):
        raise NotImplementedError()

    def _reward(self, action):
        raise NotImplementedError()

    def calc_rewards(self, a, state):
        print("Please do not directly instantiate BaseRobotEnv")
        raise NotImplementedError()

    def get_eye_pos_orientation(self):
        """Used in CameraEnv.setup"""
        eye_pos = self.robot.eyes.get_position()
        x, y, z ,w = self.robot.eyes.get_orientation()
        eye_quat = [w, x, y, z]
        return eye_pos, eye_quat

    def move_robot(self, init_x, init_y, init_z):
        "Used by multiplayer building to move sideways, to another running lane."
        self.cpp_robot.query_position()
        pose = self.cpp_robot.root_part.pose()
        pose.move_xyz(init_x, init_y, init_z)  # Works because robot loads around (0,0,0), and some robots have z != 0 that is left intact
        self.cpp_robot.set_pose(pose)

    def camera_adjust(self):
        x, y, z = self.body_xyz
        self.camera_x = 0.98*self.camera_x + (1-0.98)*x
        self.camera.move_and_look_at(self.camera_x, y-2.0, 1.4, x, y, 1.0)

    def find_best_k_views(self, eye_pos, all_dist, all_pos, avoid_block=False):
        least_order = (np.argsort(all_dist))
        top_k = []
        num_to_test = self.k * 2
        curr_num = 0
        all_pos = np.array(all_pos)

        if not avoid_block:
            return least_order[:self.k]

        while len(top_k) < self.k:
            curr_order = least_order[curr_num: curr_num + num_to_test]
            curr_pos =  all_pos[curr_order]
            print("Curr num", curr_num, "top k", len(top_k), self.k)
            if len(curr_pos) <= p.MAX_RAY_INTERSECTION_BATCH_SIZE:
                collisions = list(p.rayTestBatch([eye_pos] * len(curr_pos), curr_pos))
            else:
                collisions = []
                curr_i = 0
                while (curr_i < len(curr_pos)):
                    curr_n = min(len(curr_pos), curr_i + p.MAX_RAY_INTERSECTION_BATCH_SIZE - 1)
                    collisions = collisions + list(p.rayTestBatch([eye_pos] * (curr_n - curr_i), curr_pos[curr_i: curr_n]))
                    curr_i = curr_n
            has_collision = [c[0] > 0 for c in collisions]
            ## (hzyjerry): ray_casting-based view selection occasionally gives unstable behaviour. Will keep watching on this
            for i, x in enumerate(curr_order):
                if not has_collision[i]:
                    top_k.append(x)
        return top_k


    def getExtendedObservation(self):
        pass

    # Gym v0.10.5 compatibility
    reset = _reset
    step  = _step


class CameraRobotEnv(BaseRobotEnv):
    """CameraRobotEnv has full modalities. If it's initialized with mode="SENSOR",
    PC renderer is not initialized to save time.
    """
    multiprocessing = True
    def __init__(self, config, gpu_idx, scene_type, tracking_camera):
        ## The following properties are already instantiated inside xxx_env.py:
        BaseRobotEnv.__init__(self, config, tracking_camera, scene_type, gpu_idx)

        if self.gui:
            self.screen_arr = np.zeros([512, 512, 3])
        
        self.test_env = "TEST_ENV" in os.environ.keys() and os.environ['TEST_ENV'] == "True"
        self._use_filler = config["use_filler"]
        self._require_camera_input = 'rgb_filled' in self.config["output"] or \
                                     'rgb_prefilled' in self.config["output"] or \
                                     'depth' in self.config["output"] or \
                                     'normal' in self.config["output"] or \
                                     'semantics' in self.config["output"]
        self._require_rgb = 'rgb_filled' in self.config["output"] or "rgb_prefilled" in self.config["output"]
        self._require_depth = 'depth' in self.config["output"]
        self._require_normal = 'depth' in self.config["output"]
        self._require_semantics = 'semantics' in self.config["output"]
        self._semantic_source = 1
        self._semantic_color = 1
        if self._require_semantics:
            assert "semantic_source" in self.config.keys(), "semantic_source not specified in configuration"
            assert "semantic_color" in self.config.keys(), "semantic_color not specified in configuration"
            assert self.config["semantic_source"] in [1, 2], "semantic_source not valid"
            assert self.config["semantic_color"] in [1, 2, 3], "semantic_source not valid"
            self._semantic_source = self.config["semantic_source"]
            self._semantic_color  = self.config["semantic_color"]
        self._require_normal = 'normal' in self.config["output"]

        #if self._require_camera_input:
        self.model_path = get_model_path(self.model_id)

        self.save_frame  = 0
        self.fps = 0


    def reset_observations(self):
        ## Initialize blank render image
        self.render_rgb_filled = np.zeros((self.windowsz, self.windowsz, 3))
        self.render_rgb_prefilled = np.zeros((self.windowsz, self.windowsz, 3))
        self.render_depth = np.zeros((self.windowsz, self.windowsz, 1))
        self.render_normal = np.zeros((self.windowsz, self.windowsz, 3))
        self.render_semantics = np.zeros((self.windowsz, self.windowsz, 3))

    def robot_introduce(self, robot):
        BaseRobotEnv.robot_introduce(self, robot)
        self.setup_rendering_camera()

    def scene_introduce(self, gravity=9.8, collision_enabled=True):
        BaseRobotEnv.scene_introduce(self, gravity=gravity, collision_enabled=collision_enabled)

    def setup_rendering_camera(self):
        if self.test_env:
            return
        self.r_camera_rgb = None     ## Rendering engine
        self.r_camera_mul = None     ## Multi channel rendering engine
        self.r_camera_dep = None
        #self.check_port_available()

        ui_map = {
            1: OneViewUI,
            2: TwoViewUI,
            3: ThreeViewUI,
            4: FourViewUI,
        }

        assert self.config["ui_num"] == len(self.config['ui_components']), "In configuration, ui_num is not equal to the number of ui components"
        if self.config["display_ui"]:
            ui_num = self.config["ui_num"]
            self.UI = ui_map[ui_num](self.windowsz, self, self.port_ui)

        if self._require_camera_input:
            self.setup_camera_multi()
            self.setup_camera_pc()

        if self.config["mode"] == "web_ui":
            ui_num = self.config["ui_num"]
            self.webUI = ui_map[ui_num](self.windowsz, self, self.port_ui, use_pygame=False)

    def _reset(self):
        self.reset_observations()
        sensor_state = BaseRobotEnv._reset(self)
        self.potential = self.robot.calc_potential()
        eye_pos, eye_quat = self.get_eye_pos_orientation()
        pose = [eye_pos, eye_quat]
        
        observations = self.render_observations(pose)
        return observations #, sensor_state


    def add_text(self, img):
        return img

    def _step(self, a):
        t = time.time()
        base_obs, sensor_reward, done, sensor_meta = BaseRobotEnv._step(self, a)
        dt = time.time() - t
        # Speed bottleneck
        observations = base_obs
        self.fps = 0.9 * self.fps + 0.1 * 1/dt

        pose = [sensor_meta['eye_pos'], sensor_meta['eye_quat']]
        sensor_meta.pop("eye_pos", None)
        sensor_meta.pop("eye_quat", None)
        #sensor_meta["sensor"] = sensor_state

        if self.gui:
            if self.config["display_ui"]:
                self.render_to_UI()
                self.save_frame += 1
            elif self._require_camera_input:
                # Use non-pygame GUI
                self.r_camera_rgb.renderToScreen()

        if self.config["mode"] == 'web_ui':
            self.render_to_webUI()

        if not self._require_camera_input or self.test_env:
            ## No camera input (rgb/depth/normal/semantics)
            return base_obs, sensor_reward, done, sensor_meta
        else:
            if self.config["show_diagnostics"] and self._require_rgb:
                self.render_rgb_filled = self.add_text(self.render_rgb_filled)

            robot_pos = self.robot.get_position()
            debugmode = 0
            if debugmode:
                print("Eye position", sensor_meta['eye_pos'])
            debugmode = 0
            if debugmode:
                print("Environment observation keys", observations.keys)
            return observations, sensor_reward, done, sensor_meta


    def render_component(self, tag):
        if tag == View.RGB_FILLED:
            return self.render_rgb_filled
        if tag == View.RGB_PREFILLED:
            return self.render_rgb_prefilled
        if tag == View.DEPTH:
            scaled_depth = self.render_depth.copy()
            scaled_depth = scaled_depth * DEPTH_SCALE_FACTOR + DEPTH_OFFSET_FACTOR
            overflow = scaled_depth > 255.
            scaled_depth[overflow] = 255.
            return scaled_depth
        if tag == View.NORMAL:
            return self.render_normal
        if tag == View.SEMANTICS:
            print("Render components: semantics", np.mean(self.render_semantics))
            return self.render_semantics
        if tag == View.PHYSICS:
            return self.render_physics()
        if tag == View.MAP:
            return self.render_map()

    def render_to_UI(self):
        '''Works for different UI: UI_SIX, UI_FOUR, UI_TWO
        '''
        if not self.config["display_ui"]:
            return

        self.UI.refresh()

        for component in self.UI.components:
            self.UI.update_view(self.render_component(component), component)

    def render_to_webUI(self):
        '''Works for different UI: UI_SIX, UI_FOUR, UI_TWO
        '''
        if not self.config["mode"] == 'web_ui':
            return

        self.webUI.refresh()

        for component in self.webUI.components:
            self.webUI.update_view(self.render_component(component), component)



    def _close(self):
        BaseEnv._close(self)

        if self._require_camera_input:
            self.r_camera_mul.terminate()
            self.r_camera_rgb._close()

            if self.r_camera_dep is not None:
                self.r_camera_dep.terminate()
            if self._require_normal:
                self.r_camera_norm.terminate()
            if self._require_semantics:
                self.r_camera_semt.terminate()

        if self.config["display_ui"]:
            self.UI._close()

    def get_key_pressed(self, relevant=None):
        pressed_keys = []
        events = p.getKeyboardEvents()
        key_codes = events.keys()
        for key in key_codes:
            pressed_keys.append(key)
        return pressed_keys

    def get_blank_visuals(self):
        return np.zeros((256, 256, 4))

    def render_observations(self, pose):
        '''Render all environment observations, called inside every step()
        Input
            @pose: current robot pose
        Return:
            @observation: dict with key values being output components user
            specified in config file
        TODO:
            @hzyjerry: add noise to observation
        '''

        self.render_nonviz_sensor = self.robot.calc_state()

        if self._require_camera_input:
            self.r_camera_rgb.setNewPose(pose)
            all_dist, all_pos = self.r_camera_rgb.getAllPoseDist(pose)
            top_k = self.find_best_k_views(pose[0], all_dist, all_pos, avoid_block=False)
            #with Profiler("Render to screen"):
            self.render_rgb_filled, self.render_depth, self.render_semantics, self.render_normal, self.render_rgb_prefilled = self.r_camera_rgb.renderOffScreen(pose, top_k, rgb=self._require_rgb)

        observations = {}
        for output in self.config["output"]:
            try:
                observations[output] = getattr(self, "render_" + output)
            except Exception as e:
                raise Exception("Output component {} is not available".format(output))

        #visuals = np.concatenate(visuals, 2)
        return observations

    def get_observations(self):
        observations = {}
        for output in self.config["output"]:
            try:
                observations[output] = getattr(self, "render_" + output)
            except Exception as e:
                raise Exception("Output component {} is not available".format(output))
        
        #visuals = np.concatenate(visuals, 2)
        return observations

    def setup_camera_pc(self):
        ## Camera specific
        assert(self._require_camera_input)
        if self.scene_type == "building":
            self.dataset = ViewDataSet3D(
                transform = np.array,
                mist_transform = np.array,
                seqlen = 2,
                off_3d = False,
                train = False,
                overwrite_fofn=True, env = self, only_load = self.config["model_id"])

        scene_dict = dict(zip(self.dataset.scenes, range(len(self.dataset.scenes))))
        ## Todo: (hzyjerry) more error handling
        if not self.model_id in scene_dict.keys():
             raise error.Error("Dataset not found: model {} cannot be loaded".format(self.model_id))
        else:
            scene_id = scene_dict[self.model_id]
        uuids, rts = self.dataset.get_scene_info(scene_id)

        targets, sources, source_depths, poses = [], [], [], []
        source_semantics = []

        if not self.multiprocessing or self.config["envname"] == "TestEnv":
            all_data = self.dataset.get_multi_index([v for k, v in uuids])
            for i, data in enumerate(all_data):
                target, target_depth = data[1], data[3]
                if not self._require_rgb:
                    continue
                ww = target.shape[0] // 8 + 2
                target[:ww, :, :] = target[ww, :, :]
                target[-ww:, :, :] = target[-ww, :, :]

                if self.scale_up !=1:
                    target = cv2.resize(
                        target,None,
                        fx=1.0/self.scale_up,
                        fy=1.0/self.scale_up,
                        interpolation = cv2.INTER_CUBIC)
                    target_depth =  cv2.resize(
                        target_depth, None,
                        fx=1.0/self.scale_up,
                        fy=1.0/self.scale_up,
                        interpolation = cv2.INTER_CUBIC)
                pose = data[-1][0].numpy()
                targets.append(target)
                poses.append(pose)
                sources.append(target)
                source_depths.append(target_depth)
        else:
            all_data = self.dataset.get_multi_index([v for k, v in uuids])
            for i, data in enumerate(all_data):
                target, target_depth = data[1], data[3]
                if not self._require_rgb:
                    continue
                ww = target.shape[0] // 8 + 2
                target[:ww, :, :] = target[ww, :, :]
                target[-ww:, :, :] = target[-ww, :, :]

                if self.scale_up !=1:

                    target = cv2.resize(
                        target,None,
                        fx=1.0/self.scale_up,
                        fy=1.0/self.scale_up,
                        interpolation = cv2.INTER_CUBIC)
                    target_depth =  cv2.resize(
                        target_depth, None,
                        fx=1.0/self.scale_up,
                        fy=1.0/self.scale_up,
                        interpolation = cv2.INTER_CUBIC)
                pose = data[-1][0].numpy()
                targets.append(target)
                poses.append(pose)
                sources.append(target)
                source_depths.append(target_depth) 
        
        self.r_camera_rgb = PCRenderer(self.port_rgb, sources, source_depths, target, rts, 
                                       scale_up=self.scale_up, 
                                       semantics=source_semantics,
                                       gui=self.gui, 
                                       use_filler=self._use_filler,  
                                       gpu_idx=self.gpu_idx,
                                       windowsz=self.windowsz, 
                                       env = self)

    def setup_camera_multi(self):
        assert(self._require_camera_input)
        def camera_multi_excepthook(exctype, value, tb):
            print("killing", self.r_camera_mul)
            self.r_camera_mul.terminate()
            if self.r_camera_dep is not None:
                self.r_camera_dep.terminate()
            if self._require_normal:
                self.r_camera_norm.terminate()
            if self._require_semantics:
                self.r_camera_semt.terminate()
            while tb:
                if exctype == KeyboardInterrupt:
                    print("Exiting Gibson...")
                    return
                filename = tb.tb_frame.f_code.co_filename
                name = tb.tb_frame.f_code.co_name
                lineno = tb.tb_lineno
                print('   File "%.500s", line %d, in %.500s' %(filename, lineno, name))
                tb = tb.tb_next
            print(' %s: %s' %(exctype.__name__, value))

        sys.excepthook = camera_multi_excepthook
        enable_render_smooth = 0

        dr_path = osp.join(osp.dirname(osp.abspath(gibson.__file__)), 'core', 'channels', 'depth_render')
        cur_path = os.getcwd()
        os.chdir(dr_path)

        render_main  = "./depth_render --GPU {} --modelpath {} -w {} -h {} -f {} -p {}".format(self.gpu_idx, self.model_path, self.windowsz, self.windowsz, self.config["fov"]/np.pi*180, self.port_depth)
        render_norm  = "./depth_render --GPU {} --modelpath {} -n 1 -w {} -h {} -f {} -p {}".format(self.gpu_idx, self.model_path, self.windowsz, self.windowsz, self.config["fov"]/np.pi*180, self.port_normal)
        render_semt  = "./depth_render --GPU {} --modelpath {} -t 1 -r {} -c {} -w {} -h {} -f {} -p {}".format(self.gpu_idx, self.model_path, self._semantic_source, self._semantic_color, self.windowsz, self.windowsz, self.config["fov"]/np.pi*180, self.port_sem)
        
        self.r_camera_mul = subprocess.Popen(shlex.split(render_main), shell=False)
        #self.r_camera_dep = subprocess.Popen(shlex.split(render_depth), shell=False)
        if self._require_normal:
            self.r_camera_norm = subprocess.Popen(shlex.split(render_norm), shell=False)
        if self._require_semantics:
            self.r_camera_semt = subprocess.Popen(shlex.split(render_semt), shell=False)

        os.chdir(cur_path)


    def check_port_available(self):
        assert(self._require_camera_input)
        # TODO (hzyjerry)
        ports = []
        if self._require_depth: ports.append(self.port_depth)
        if self._require_normal: ports.append(self.port_normal)
        if self._require_semantics: ports.append(self.port_sem)
        for port in ports:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                result = s.bind(("127.0.0.1", port - 1))
            except socket.error as e:
                raise e
                raise error.Error("Gibson initialization Error: port {} is in use".format(port))
    # Gym v0.10.5 compatibility
    reset = _reset
    step  = _step
    close = _close


class SemanticRobotEnv(CameraRobotEnv):
    def __init__(self, config, gpu_idx, scene_type, tracking_camera):
        CameraRobotEnv.__init__(self, config, gpu_idx, scene_type, tracking_camera)

    def robot_introduce(self, robot):
        CameraRobotEnv.robot_introduce(self, robot)
        self.setup_semantic_parser()

    def setup_semantic_parser(self):
        #assert('semantics' in self.config["output"])
        def semantic_excepthook(exctype, value, tb):
            print("killing", self.r_camera_mul)
            self.r_camera_mul.terminate()
            if self.r_camera_dep:
                self.r_camera_dep.terminate()
            if self._require_normal:
                self.r_camera_norm.terminate()
            if self._require_semantics:
                self.r_camera_semt.terminate()
            while tb:
                if exctype == KeyboardInterrupt:
                    print("Exiting Gibson...")
                    return
                filename = tb.tb_frame.f_code.co_filename
                name = tb.tb_frame.f_code.co_name
                lineno = tb.tb_lineno
                print('   File "%.500s", line %d, in %.500s' %(filename, lineno, name))
                tb = tb.tb_next
            print(' %s: %s' %(exctype.__name__, value))

        sys.excepthook = semantic_excepthook
        dr_path = osp.join(osp.dirname(osp.abspath(gibson.__file__)), 'core', 'channels', 'depth_render')
        cur_path = os.getcwd()
        os.chdir(dr_path)
        load_semantic  = "./semantic --modelpath {} -r {} ".format(self.model_path, self._semantic_source)
        self.semantic_server = subprocess.Popen(shlex.split(load_semantic), shell=False)
        os.chdir(cur_path)

        self._context_sem = zmq.Context()
        self.semantic_client = self._context_sem.socket(zmq.REQ)
        self.semantic_client.connect("tcp://localhost:{}".format(5055))

        self.semantic_client.send_string("Ready")
        semantic_msg = self.semantic_client.recv()
        self.semantic_pos = np.frombuffer(semantic_msg, dtype=np.float32).reshape((-1, 3))

        if self._semantic_source == 2:
            _, semantic_ids, _ = get_segmentId_by_name_MP3D(osp.join(self.model_path, "semantic.house"), "chair")
        elif self._semantic_source == 1:
            _, semantic_ids, _ = get_segmentId_by_name_2D3DS(osp.join(self.model_path, "semantic.mtl"), osp.join(self.model_path, "semantic.obj"), "chair")

        self.semantic_pos = self.semantic_pos[semantic_ids, :]

        debugmode=0
        if debugmode:
            self.semantic_pos = np.array([[0, 0, 0.2]])
    
    def dist_to_semantic_pos(self):
        pos = self.robot.get_position()
        x, y, z, w = self.robot.get_orientation()
        #print(self.semantic_pos)
        #print(pos, orn)

        diff_pos = self.semantic_pos - pos
        dist_to_robot = np.sqrt(np.sum(diff_pos * diff_pos, axis = 1))
        diff_unit = (diff_pos.T / dist_to_robot).T

        #TODO: (hzyjerry) orientation is still buggy
        orn_unit = quat2mat([w, x, y, z]).dot(np.array([-1, 0, 0]))
        orn_to_robot = np.arccos(diff_unit.dot(orn_unit))
        return dist_to_robot, orn_to_robot

    def get_close_semantic_pos(self, dist_max=1.0, orn_max=np.pi/5):
        '''Find the index of semantic positions close to the agent, within max
        distance and max orientation
        Return: list of index of the semantic positions, corresponding the index
            in self.semantic_pos
        '''
        dists, orns = self.dist_to_semantic_pos()
        return [i for i in range(self.semantic_pos.shape[0]) if dists[i] < dist_max and orns[i] < orn_max]

    def step(self, action, tag=True):
        #self.close_semantic_ids = self.get_close_semantic_pos()
        return CameraRobotEnv.step(self, action)


STR_TO_PYGAME_KEY = {
    'a': pygame.K_a,
    'b': pygame.K_b,
    'c': pygame.K_c,
    'd': pygame.K_d,
    'e': pygame.K_e,
    'f': pygame.K_f,
    'g': pygame.K_g,
    'h': pygame.K_h,
    'i': pygame.K_i,
    'j': pygame.K_j,
    'k': pygame.K_k,
    'l': pygame.K_l,
    'm': pygame.K_m,
    'n': pygame.K_n,
    'o': pygame.K_o,
    'p': pygame.K_p,
    'q': pygame.K_q,
    'r': pygame.K_r,
    's': pygame.K_s,
    't': pygame.K_t,
    'u': pygame.K_u,
    'v': pygame.K_v,
    'w': pygame.K_w,
    'x': pygame.K_x,
    'y': pygame.K_y,
    'z': pygame.K_z,
}



#-*- coding: UTF-8 -*-

"""
Some fancy helper functions.
"""

import os
import ctypes
import operator

from distutils.sysconfig import get_python_lib
import re
import sys

try: import numpy
except ImportError: numpy = None

import logging;logger = logging.getLogger("pyassimp")

from .errors import AssimpError

additional_dirs, ext_whitelist = [],[]

# populate search directories and lists of allowed file extensions
# depending on the platform we're running on.
if os.name=='posix':
    additional_dirs.append('./')
    additional_dirs.append('/usr/lib/')
    additional_dirs.append('/usr/lib/x86_64-linux-gnu/')
    additional_dirs.append('/usr/local/lib/')

    if 'LD_LIBRARY_PATH' in os.environ:
        additional_dirs.extend([item for item in os.environ['LD_LIBRARY_PATH'].split(':') if item])

    # check if running from anaconda.
    if "conda" or "continuum" in sys.version.lower():
      cur_path = get_python_lib()
      pattern = re.compile('.*\/lib\/')
      conda_lib = pattern.match(cur_path).group()
      logger.info("Adding Anaconda lib path:"+ conda_lib)
      additional_dirs.append(conda_lib)

    # note - this won't catch libassimp.so.N.n, but
    # currently there's always a symlink called
    # libassimp.so in /usr/local/lib.
    ext_whitelist.append('.so')
    # libassimp.dylib in /usr/local/lib
    ext_whitelist.append('.dylib')

elif os.name=='nt':
    ext_whitelist.append('.dll')
    path_dirs = os.environ['PATH'].split(';')
    additional_dirs.extend(path_dirs)

def vec2tuple(x):
    """ Converts a VECTOR3D to a Tuple """
    return (x.x, x.y, x.z)

def transform(vector3, matrix4x4):
    """ Apply a transformation matrix on a 3D vector.

    :param vector3: array with 3 elements
    :param matrix4x4: 4x4 matrix
    """
    if numpy:
        return numpy.dot(matrix4x4, numpy.append(vector3, 1.))
    else:
        m0,m1,m2,m3 = matrix4x4; x,y,z = vector3
        return [
            m0[0]*x + m0[1]*y + m0[2]*z + m0[3],
            m1[0]*x + m1[1]*y + m1[2]*z + m1[3],
            m2[0]*x + m2[1]*y + m2[2]*z + m2[3],
            m3[0]*x + m3[1]*y + m3[2]*z + m3[3]
            ]

def _inv(matrix4x4):
    m0,m1,m2,m3 = matrix4x4

    det  =  m0[3]*m1[2]*m2[1]*m3[0] - m0[2]*m1[3]*m2[1]*m3[0] - \
            m0[3]*m1[1]*m2[2]*m3[0] + m0[1]*m1[3]*m2[2]*m3[0] + \
            m0[2]*m1[1]*m2[3]*m3[0] - m0[1]*m1[2]*m2[3]*m3[0] - \
            m0[3]*m1[2]*m2[0]*m3[1] + m0[2]*m1[3]*m2[0]*m3[1] + \
            m0[3]*m1[0]*m2[2]*m3[1] - m0[0]*m1[3]*m2[2]*m3[1] - \
            m0[2]*m1[0]*m2[3]*m3[1] + m0[0]*m1[2]*m2[3]*m3[1] + \
            m0[3]*m1[1]*m2[0]*m3[2] - m0[1]*m1[3]*m2[0]*m3[2] - \
            m0[3]*m1[0]*m2[1]*m3[2] + m0[0]*m1[3]*m2[1]*m3[2] + \
            m0[1]*m1[0]*m2[3]*m3[2] - m0[0]*m1[1]*m2[3]*m3[2] - \
            m0[2]*m1[1]*m2[0]*m3[3] + m0[1]*m1[2]*m2[0]*m3[3] + \
            m0[2]*m1[0]*m2[1]*m3[3] - m0[0]*m1[2]*m2[1]*m3[3] - \
            m0[1]*m1[0]*m2[2]*m3[3] + m0[0]*m1[1]*m2[2]*m3[3]

    return[[( m1[2]*m2[3]*m3[1] - m1[3]*m2[2]*m3[1] + m1[3]*m2[1]*m3[2] - m1[1]*m2[3]*m3[2] - m1[2]*m2[1]*m3[3] + m1[1]*m2[2]*m3[3]) /det,
            ( m0[3]*m2[2]*m3[1] - m0[2]*m2[3]*m3[1] - m0[3]*m2[1]*m3[2] + m0[1]*m2[3]*m3[2] + m0[2]*m2[1]*m3[3] - m0[1]*m2[2]*m3[3]) /det,
            ( m0[2]*m1[3]*m3[1] - m0[3]*m1[2]*m3[1] + m0[3]*m1[1]*m3[2] - m0[1]*m1[3]*m3[2] - m0[2]*m1[1]*m3[3] + m0[1]*m1[2]*m3[3]) /det,
            ( m0[3]*m1[2]*m2[1] - m0[2]*m1[3]*m2[1] - m0[3]*m1[1]*m2[2] + m0[1]*m1[3]*m2[2] + m0[2]*m1[1]*m2[3] - m0[1]*m1[2]*m2[3]) /det],
           [( m1[3]*m2[2]*m3[0] - m1[2]*m2[3]*m3[0] - m1[3]*m2[0]*m3[2] + m1[0]*m2[3]*m3[2] + m1[2]*m2[0]*m3[3] - m1[0]*m2[2]*m3[3]) /det,
            ( m0[2]*m2[3]*m3[0] - m0[3]*m2[2]*m3[0] + m0[3]*m2[0]*m3[2] - m0[0]*m2[3]*m3[2] - m0[2]*m2[0]*m3[3] + m0[0]*m2[2]*m3[3]) /det,
            ( m0[3]*m1[2]*m3[0] - m0[2]*m1[3]*m3[0] - m0[3]*m1[0]*m3[2] + m0[0]*m1[3]*m3[2] + m0[2]*m1[0]*m3[3] - m0[0]*m1[2]*m3[3]) /det,
            ( m0[2]*m1[3]*m2[0] - m0[3]*m1[2]*m2[0] + m0[3]*m1[0]*m2[2] - m0[0]*m1[3]*m2[2] - m0[2]*m1[0]*m2[3] + m0[0]*m1[2]*m2[3]) /det],
           [( m1[1]*m2[3]*m3[0] - m1[3]*m2[1]*m3[0] + m1[3]*m2[0]*m3[1] - m1[0]*m2[3]*m3[1] - m1[1]*m2[0]*m3[3] + m1[0]*m2[1]*m3[3]) /det,
            ( m0[3]*m2[1]*m3[0] - m0[1]*m2[3]*m3[0] - m0[3]*m2[0]*m3[1] + m0[0]*m2[3]*m3[1] + m0[1]*m2[0]*m3[3] - m0[0]*m2[1]*m3[3]) /det,
            ( m0[1]*m1[3]*m3[0] - m0[3]*m1[1]*m3[0] + m0[3]*m1[0]*m3[1] - m0[0]*m1[3]*m3[1] - m0[1]*m1[0]*m3[3] + m0[0]*m1[1]*m3[3]) /det,
            ( m0[3]*m1[1]*m2[0] - m0[1]*m1[3]*m2[0] - m0[3]*m1[0]*m2[1] + m0[0]*m1[3]*m2[1] + m0[1]*m1[0]*m2[3] - m0[0]*m1[1]*m2[3]) /det],
           [( m1[2]*m2[1]*m3[0] - m1[1]*m2[2]*m3[0] - m1[2]*m2[0]*m3[1] + m1[0]*m2[2]*m3[1] + m1[1]*m2[0]*m3[2] - m1[0]*m2[1]*m3[2]) /det,
            ( m0[1]*m2[2]*m3[0] - m0[2]*m2[1]*m3[0] + m0[2]*m2[0]*m3[1] - m0[0]*m2[2]*m3[1] - m0[1]*m2[0]*m3[2] + m0[0]*m2[1]*m3[2]) /det,
            ( m0[2]*m1[1]*m3[0] - m0[1]*m1[2]*m3[0] - m0[2]*m1[0]*m3[1] + m0[0]*m1[2]*m3[1] + m0[1]*m1[0]*m3[2] - m0[0]*m1[1]*m3[2]) /det,
            ( m0[1]*m1[2]*m2[0] - m0[2]*m1[1]*m2[0] + m0[2]*m1[0]*m2[1] - m0[0]*m1[2]*m2[1] - m0[1]*m1[0]*m2[2] + m0[0]*m1[1]*m2[2]) /det]]

def get_bounding_box(scene):
    bb_min = [1e10, 1e10, 1e10] # x,y,z
    bb_max = [-1e10, -1e10, -1e10] # x,y,z
    inv = numpy.linalg.inv if numpy else _inv
    return get_bounding_box_for_node(scene.rootnode, bb_min, bb_max, inv(scene.rootnode.transformation))

def get_bounding_box_for_node(node, bb_min, bb_max, transformation):

    if numpy:
        transformation = numpy.dot(transformation, node.transformation)
    else:
        t0,t1,t2,t3 = transformation
        T0,T1,T2,T3 = node.transformation
        transformation = [ [
                t0[0]*T0[0] + t0[1]*T1[0] + t0[2]*T2[0] + t0[3]*T3[0],
                t0[0]*T0[1] + t0[1]*T1[1] + t0[2]*T2[1] + t0[3]*T3[1],
                t0[0]*T0[2] + t0[1]*T1[2] + t0[2]*T2[2] + t0[3]*T3[2],
                t0[0]*T0[3] + t0[1]*T1[3] + t0[2]*T2[3] + t0[3]*T3[3]
            ],[
                t1[0]*T0[0] + t1[1]*T1[0] + t1[2]*T2[0] + t1[3]*T3[0],
                t1[0]*T0[1] + t1[1]*T1[1] + t1[2]*T2[1] + t1[3]*T3[1],
                t1[0]*T0[2] + t1[1]*T1[2] + t1[2]*T2[2] + t1[3]*T3[2],
                t1[0]*T0[3] + t1[1]*T1[3] + t1[2]*T2[3] + t1[3]*T3[3]
            ],[
                t2[0]*T0[0] + t2[1]*T1[0] + t2[2]*T2[0] + t2[3]*T3[0],
                t2[0]*T0[1] + t2[1]*T1[1] + t2[2]*T2[1] + t2[3]*T3[1],
                t2[0]*T0[2] + t2[1]*T1[2] + t2[2]*T2[2] + t2[3]*T3[2],
                t2[0]*T0[3] + t2[1]*T1[3] + t2[2]*T2[3] + t2[3]*T3[3]
            ],[
                t3[0]*T0[0] + t3[1]*T1[0] + t3[2]*T2[0] + t3[3]*T3[0],
                t3[0]*T0[1] + t3[1]*T1[1] + t3[2]*T2[1] + t3[3]*T3[1],
                t3[0]*T0[2] + t3[1]*T1[2] + t3[2]*T2[2] + t3[3]*T3[2],
                t3[0]*T0[3] + t3[1]*T1[3] + t3[2]*T2[3] + t3[3]*T3[3]
            ] ]

    for mesh in node.meshes:
        for v in mesh.vertices:
            v = transform(v, transformation)
            bb_min[0] = min(bb_min[0], v[0])
            bb_min[1] = min(bb_min[1], v[1])
            bb_min[2] = min(bb_min[2], v[2])
            bb_max[0] = max(bb_max[0], v[0])
            bb_max[1] = max(bb_max[1], v[1])
            bb_max[2] = max(bb_max[2], v[2])


    for child in node.children:
        bb_min, bb_max = get_bounding_box_for_node(child, bb_min, bb_max, transformation)

    return bb_min, bb_max

def try_load_functions(library_path, dll):
    '''
    Try to bind to aiImportFile and aiReleaseImport

    Arguments
    ---------
    library_path: path to current lib
    dll:          ctypes handle to library

    Returns
    ---------
    If unsuccessful:
        None
    If successful:
        Tuple containing (library_path,
                          load from filename function,
                          load from memory function,
                          export to filename function,
                          export to blob function,
                          release function,
                          ctypes handle to assimp library)
    '''

    try:
        load     = dll.aiImportFile
        release  = dll.aiReleaseImport
        load_mem = dll.aiImportFileFromMemory
        export   = dll.aiExportScene
        export2blob = dll.aiExportSceneToBlob
    except AttributeError:
        #OK, this is a library, but it doesn't have the functions we need
        return None

    # library found!
    from .structs import Scene, ExportDataBlob
    load.restype = ctypes.POINTER(Scene)
    load_mem.restype = ctypes.POINTER(Scene)
    export2blob.restype = ctypes.POINTER(ExportDataBlob)
    return (library_path, load, load_mem, export, export2blob, release, dll)

def search_library():
    '''
    Loads the assimp library.
    Throws exception AssimpError if no library_path is found

    Returns: tuple, (load from filename function,
                     load from memory function,
                     export to filename function,
                     export to blob function,
                     release function,
                     dll)
    '''
    #this path
    folder = os.path.dirname(__file__)

    # silence 'DLL not found' message boxes on win
    try:
        ctypes.windll.kernel32.SetErrorMode(0x8007)
    except AttributeError:
        pass

    candidates = []
    # test every file
    for curfolder in [folder]+additional_dirs:
        if os.path.isdir(curfolder):
            for filename in os.listdir(curfolder):
                # our minimum requirement for candidates is that
                # they should contain 'assimp' somewhere in
                # their name                                  
                if filename.lower().find('assimp')==-1 : 
                    continue
                is_out=1
                for et in ext_whitelist:
                  if et in filename.lower():
                    is_out=0
                    break
                if is_out:
                  continue
                
                library_path = os.path.join(curfolder, filename)
                logger.debug('Try ' + library_path)
                try:
                    dll = ctypes.cdll.LoadLibrary(library_path)
                except Exception as e:
                    logger.warning(str(e))
                    # OK, this except is evil. But different OSs will throw different
                    # errors. So just ignore any errors.
                    continue
                # see if the functions we need are in the dll
                loaded = try_load_functions(library_path, dll)
                if loaded: candidates.append(loaded)

    if not candidates:
        # no library found
        raise AssimpError("assimp library not found")
    else:
        # get the newest library_path
        candidates = map(lambda x: (os.lstat(x[0])[-2], x), candidates)
        res = max(candidates, key=operator.itemgetter(0))[1]
        logger.debug('Using assimp library located at ' + res[0])

        # XXX: if there are 1000 dll/so files containing 'assimp'
        # in their name, do we have all of them in our address
        # space now until gc kicks in?

        # XXX: take version postfix of the .so on linux?
        return res[1:]

def hasattr_silent(object, name):
    """
        Calls hasttr() with the given parameters and preserves the legacy (pre-Python 3.2)
        functionality of silently catching exceptions.

        Returns the result of hasatter() or False if an exception was raised.
    """

    try:
        if not object:
            return False
        return hasattr(object, name)
    except AttributeError:
        return False
